/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityDsp.cpp								*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/
  
#include "EntityDsp.h"
#include "dsp/dsp_code.h"

#include <fastmath.h>

//#define DEBUG_DSP

EntityDsp::EntityDsp(ParaStructure *paras_) {

	paras = paras_;
	
	if(!(dspHandle = dsp_init(1, dspCode)))
		TRACE("ERROR: Cannot initialise DSP\n");
	
	soundBuffer = new int [paras->getnValues()];

	rtems_task_wake_after(3); // Give DSP a bit of time

	waitForDsp();
	
	TRACE("Beginning transfer of LUT initialisation data\n");

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
	// Transfer LUT initialisation data
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

	// 2^23 =  8388607.0
	// 2^24 = 16777216.0
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
	// 1. Pitch to frequency (for audio oscillators & filters)
	//
	// 			     2^24 * frequency
	//  per-sample-increment =   ----------------
	//    			   	sampleRate
	//
	// The middle of the LUT has pitch = 0, middle C (60)
	// which has frequency = 261.6255653 Hz (A=440)
	//
	// The LUT passed to the DSP is in two parts: coarse 
	// and fine, each 256 words in length. The coarse
	// resolution is 256/400 semitones; the fine resolution
	// is 1/400 semitones ( cent).
	//
	// In the 56K, the following equation is used:
	//
	// freq = P2Fc[pitch>>8]
	// freq += freq * P2Ff[pitch&0xff]
	//
	// Where pitch is a 16bit number, in  cent units,
	// with middle C (60) = 32768, and A440 = 36368.
	//
	// To maximise precision, the integers are unsigned.

	TRACE("1. Pitch to Frequency\n");

	// 1a. Coarse P2F LUT	

	double P2F[256];
	double max = -48000.0;
	
	for(int x = 0; x < 256; x++) {

		double pitch = (double)(x - 128) * (256.0 / 400.0);
		double freq = 261.6255653 * pow(2.0, pitch / 12.0);
		double fc = freq / 48000.0;
	
		max = fc > max ? fc : max;

		P2F[x] = fc;
	}

	/* Optimise frequency precision to maximum possible for 24 bits */

	double popt = 1; 
	
	while(max < 1.0) { max *= 2.0; popt *= 2.0; }
	while(max >= 1.0) { max *= 0.5; popt *= 0.5; }
	
	//TRACE("INFO: Using P2F optimisation factor of %f (max freq = %f)\n", popt, max);

	int P2Fc[256];

	for(int x = 0; x < 256; x++) P2Fc[x] = (int)(P2F[x] * popt * 16777216.0 + 0.5);
	
	writeDataBlockY(DSPY_P2Fc, 256, P2Fc);

	// 1b. Fine P2F LUT
	
	int P2Ff[256];
	
	for(int x = 0; x < 256; x++) {

		double ffreq = pow(2.0, (double) x / (12.0 * 400.0));
		ffreq -= 1.0; // keep only fractional offset
	
		P2Ff[x] = (int)(16777216.0 * ffreq + 0.5);
		if(P2Ff[x] >= 16777216) P2Ff[x] = 16777215;
	}
		
	writeDataBlockY(DSPY_P2Ff, 256, P2Ff);

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// 2. Frequency to Integrator feedback (for cascade filters)
	// 
	// The amount of feedback for a 6dB stage in a corner-peaking filter
	// is given by the equation:
	//
	//		  (-2.PI(Fc))
	//        K  = 	e
	//	
	//   where Fc = F/Fs, and can be the output from P2F.
	// 
	// The function is implemented as 3 LUTs, each addressed by eight different bits 
	// from the 24-bit Fc value:
	//
	//         (-S(a+b+c))           (-S.a)	           (-S.b)	     (-S.c)
	//        e		 = 	e	  .	  e	    .       e 
	//	
	//   where a is the MS 8 bits, c the LS 8 bits, b the middle 8 bits from Fc, S = 2PI
	//
	//  The three part functions have different numeric ranges. The a function
	//  has a relatively broad range, between zero and one. The b and c functions
	//  have much smaller ranges, approaching unity, ie 0.998 to 0.99999. To
	//  maximise accuracy, the values stored in the b and c LUTs are subtracted
	//  from one and then block-normalised. When applied in the DSP, the following 
	//  calculation (or its equivalent) is used:
	//
	//   a = $FF & (Fc >> 16)
	//   b = $FF & (Fc >> 8)
	//   c = $FF & Fc
	//
	//   K = F2Ia[a] * (1 - F2Ib[b]) * (1 - F2Ic[c])
	//
	//   I = F2Ia[a]			
	//   J = I - (I.F2Ib[b] >> F2IbShift)
	//   K = J - (J.F2Ic[c] >> F2IcShift)
	//
	//   where K is the feedback and F2IbShift and F2IcShift
	//   are the block-normalising shifts for F2Ib and F2Ic.

	TRACE("2. Frequency to Integrator feedback\n");

	const double S = 2 * M_PI;

	double 	F2I[256];
	int 	f2i[256];

	// 2a. Coarse F2I LUT

	for(int x = 0; x < 256; x++) {	

		double a = (double) x / 256.0;			// 0 < Fc < 1.0 (nyquist = 1.0)
		double K = pow(M_E, -S * a);			// e^(-S.a)

		f2i[x] = (int)(K * 16777216.0 + 0.5);		// (unsigned fraction)
		if(f2i[x] >= 16777216) f2i[x] = 16777215;
		
		//TRACE("a = %f\tF2Ia[%i] = %f\t ($%.6X)\n", a, x, K, f2i[x]);
	}

	writeDataBlockY(DSPY_F2Ia, 256, f2i);

	// 2b. Medium F2I LUT

	max = 0;
	
	for(int x = 0; x < 256; x++) {	

		double b = (float) x / (256.0 * 256.0);	
		double K = pow(M_E, -S * b);			// e^(-S.b), range 0.98 to 0.9999
		
		K = 1.0 - K;					// range 0.02 to 0.0001 (for example)
		if(K > max) max = K;				// record maximum

		F2I[x] = K;

		//TRACE("b = %f	F2Ib[%i] = %f\n", b, x, K);
	}

	/* Shift F2Ib values left to maximise precision */

	int shift = 0; double sf = 1.0;
	
	while(max < 0.5) { max *= 2.0; sf *= 2.0; shift++; }	// 0.5 < maximum < 1.0
	
	//TRACE("INFO: F2Ib LUT has a block-normalisation shift of %i bits\n", shift);

	for(int x = 0; x < 256; x++) {

		f2i[x] = (int)(F2I[x] * sf * 16777216.0 + 0.5);	// (unsigned fractions)

		//TRACE("F2Ib[%i] = $%.6X\n", x, f2i[x]);
		if(f2i[x] >= 16777216) TRACE("OVERFLOW: F2Ib[%i] >= 16777216\n", x);
	}

	writeDataBlockY(DSPY_F2Ib, 256, f2i);
	writeDataBlockY(DSPY_F2IbShift, 1, &shift);

	// 2c. Fine F2I LUT

	max = 0;
	
	for(int x = 0; x < 256; x++) {	

		double c = (float) x / (256.0 * 256.0 * 256.0);	
		double K = pow(M_E, -S * c);			// e^(-S.c), range 0.99998 to 0.9999999
		
		K = 1.0 - K;					// range 0.00002 to 0.0000001 (for example)
		if(K > max) max = K;				// record maximum

		F2I[x] = K;

		//TRACE("c = %f	F2Ic[%i] = %f\n", c, x, K);
	}

	/* Shift F2Ic values left to maximise precision */

	shift = 0; sf = 1.0;
	
	while(max < 0.5) { max *= 2.0; sf *= 2.0; shift++; }	// 0.5 < maximum < 1.0
	
	//TRACE("INFO: F2Ic LUT has a block-normalisation shift of %i bits\n", shift);

	for(int x = 0; x < 256; x++) {

		f2i[x] = (int)(F2I[x] * sf * 16777216.0 + 0.5);	// (unsigned fractions)

		//TRACE("F2Ic[%i] = $%.6X\n", x, f2i[x]);
		if(f2i[x] >= 16777216) TRACE("OVERFLOW: F2Ib[%i] >= 16777216\n", x);
	}

	writeDataBlockY(DSPY_F2Ic, 256, f2i);
	writeDataBlockY(DSPY_F2IcShift, 1, &shift);

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
	// 3. Filter resonance LUTs (see dsp/filter.txt)

	// These values were derived from manual calibration of the 
	// onset of the uncompensated filters. They are in the range
	// 0..8000, corresponding to fractions in the range 0..1.

	TRACE("3. Filter resonance\n");

	float res12dB[] = {

		2645, 2645, 2645, 2645, 2645, 2645, 2645, 2645,
		2645, 2644, 2643, 2642, 2641, 2636, 2632, 2627,
		2620, 2611, 2599, 2582, 2560, 2531, 2493, 2441,		
		2373, 2283, 2174, 2098, 2124, 2124, 2124, 2124,
		2124
	};

	float resScale = 8000.0;
	
	int res12dB257[257];

	// Interpolate using a cubic spline (P469, MAMP)

	for(int xd = 0; xd <= 256; xd++) { 	// address of destination

		int 	i;
		float 	Yu, Xu;       

		i = (xd >> 3) - 1;
		i = i < 0 ? 0 : i;
		i = i > 29 ? 29 : i;

		Xu = (float) xd / 8.0 - (float) i;

		const float X0 = 0.0, X1 = 1.0, X2 = 2.0, X3 = 3.0;
		const float K16 = 1.0/6.0;
		
		Yu = 	- K16 * res12dB[i+0] * (Xu - X1) * (Xu - X2) * (Xu - X3) 
		     	+ 0.5 * res12dB[i+1] * (Xu - X0) * (Xu - X2) * (Xu - X3) 
			- 0.5 * res12dB[i+2] * (Xu - X0) * (Xu - X1) * (Xu - X3) 
			+ K16 * res12dB[i+3] * (Xu - X0) * (Xu - X1) * (Xu - X2);

		Yu /= resScale;		// Divide by fsd of calibration scale
		Yu *= 0.5 / 0.7;	// Self oscillation starts at 0.7
		
		res12dB257[xd] = Yu >= 1.0 ? 8388607 : (int)(8388608.0 * Yu + 0.5);
		//TRACE("res12dB257[%i] = $%.6X\n", xd, res12dB257[xd]);
	}
	
	writeDataBlockY(DSPY_res12dB257, 257, res12dB257);

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
	// 4. Filter pitch correction LUTs (see dsp/filter.txt)

	// These values are in semitones. They represent the offset 
	// that should be added to pitch values before pitch is 
	// translated into filter coefficients.

	TRACE("4. Filter pitch correction\n");

	float fpc12dB[] = {

		-9.01, -9.01, -9.01, -9.01, -9.01, -9.01, -9.01, -9.01,
		-9.01, -9.01, -9.01, -9.02, -9.03, -9.05, -9.06, -9.08,
		-9.09, -9.08, -9.02, -8.88, -8.55, -8.08, -7.43, -6.50,
		-5.13, -3.02, +0.61, +9.38, 20.00, 20.00, 20.00, 20.00, 
		20.00
	};

	int fpc12dB257[257];

	// Interpolate using a cubic spline (P469, MAMP)

	for(int xd = 0; xd <= 256; xd++) { 	// address of destination

		int 	i;
		float 	Yu, Xu;       

		i = (xd >> 3) - 1;
		i = i < 0 ? 0 : i;
		i = i > 29 ? 29 : i;

		Xu = (float) xd / 8.0 - (float) i;

		const float X0 = 0.0, X1 = 1.0, X2 = 2.0, X3 = 3.0;
		const float K16 = 1.0/6.0;
		
		Yu = 	- K16 * fpc12dB[i+0] * (Xu - X1) * (Xu - X2) * (Xu - X3) 
		     	+ 0.5 * fpc12dB[i+1] * (Xu - X0) * (Xu - X2) * (Xu - X3) 
			- 0.5 * fpc12dB[i+2] * (Xu - X0) * (Xu - X1) * (Xu - X3) 
			+ K16 * fpc12dB[i+3] * (Xu - X0) * (Xu - X1) * (Xu - X2);

		Yu *= 400 * 256;		// Translate into cents * 256
		
		fpc12dB257[xd] = (int)(Yu + 0.5);
		//TRACE("fpc12dB257[%i] = $%.6X\n", xd, fpc12dB257[xd]);
	}
	
	writeDataBlockY(DSPY_fpc12dB257, 257, fpc12dB257);

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

	TRACE("LUTs transferred, waiting for DSP\n");

	readyForDsp();		// Tell DSP CF is ready for it to continue
	waitForDsp();		// Wait while DSP constructs LUTs
}

EntityDsp::~EntityDsp() {

	delete [] soundBuffer;
}

void EntityDsp::readyForDsp() {

	/* Tell DSP to proceed */
	command(DSPP_HV_READY);
}

void EntityDsp::inhibitDsp() {

	/* Tell DSP to clear ready flag */
	command(DSPP_HV_INHIBIT);
}

void EntityDsp::waitForDsp() {

	rtems_boolean ready = false;

	/* Wait for DSP to become ready */
	while(!ready) dsp_read_flag2(dspHandle, &ready);
}

void EntityDsp::minimiseDsp() {

	command(DSPP_HV_MINIMISE_DSP);
	rtems_boolean dspMinimised = false;
	while(!dspMinimised) dsp_read_flag3(dspHandle, &dspMinimised);
}

void EntityDsp::maximiseDsp() {

	command(DSPP_HV_MAXIMISE_DSP);
	rtems_boolean dspMinimised = true;
	while(dspMinimised) dsp_read_flag3(dspHandle, &dspMinimised);
}

void EntityDsp::update(int hCommand, int data) {

	writeData(data);
	command(hCommand);
}

void EntityDsp::update(EntitySound *pendingSound) {

	#ifdef DEBUG_DSP
	TRACE("Preparing to transfer sound data\n");	
	#endif

	int i = 0;
	
	for(int p = 0; p < paras->getnDspParas(); p++) {
		Parameter *para = paras->getDspPara(p); // get paras in dsp order
		for(int source = -1; source < paras->getnMatrixSources(); source++) {
			soundBuffer[i++] = pendingSound->getTranslatedValue(para, source);
		}
	}

	// Now we wait for the DSP to minimise processing (by suspending
	// parameter generation, zeroing interpolated parameter deltas,
	// and then generating audio as usual). The reason for this is 
	// that the DMA transfer of a sound causes a lot of DSP cycles
	// to be used up, mainly by adding wait states when swapping 
	// instructions to cache. By disabling parameter generation,
	// not only do we make more instructions cycles available;
	// we also eliminate cache swapping, as the minimised code
	// cycle (including sample generation) all fits into cache.
	
	minimiseDsp();	

	writeDataBlockY(DSPY_soundTransferBuffer, i, soundBuffer);

	maximiseDsp();

	// Cause edits to modify pending sound until next note is played

	command(DSPP_HV_SOUND_PENDING);
}

void EntityDsp::writeDataBlockX(int addr, int size, int *data) {

	defineBlock(addr, size);
	command(DSPP_HV_TRANSFER_XDATA);
	writeData(data, size);
}

void EntityDsp::writeDataBlockY(int addr, int size, int *data) {

	defineBlock(addr, size);
	command(DSPP_HV_TRANSFER_YDATA);
	writeData(data, size);
}

void EntityDsp::writeDataBlockP(int addr, int size, int *data) {

	defineBlock(addr, size);
	command(DSPP_HV_TRANSFER_PDATA);
	writeData(data, size);
}

void EntityDsp::defineBlock(int addr, int size) {

	writeData(size-1);
	command(DSPP_HV_TRANSFER_SIZEM1);
	writeData(addr);
	command(DSPP_HV_TRANSFER_ADDRESS);
}

void EntityDsp::writeData(int *data, int nWords) {

	if(!dsp_write_data(dspHandle, data, nWords))
		TRACE("ERROR: Could not write to DSP\n");

	#ifdef DEBUG_DSP
	else	TRACE("OKAY: DSP write successful\n");
	#endif
}

void EntityDsp::writeData(int data) {

	if(!dsp_write_data(dspHandle, &data, 1))
		TRACE("ERROR: Could not write to DSP\n");

	#ifdef DEBUG_DSP
	else	TRACE("OKAY: DSP write successful\n");
	#endif
}

int EntityDsp::readTotalMips() {

	command(DSPP_HV_READ_TOTAL_MIPS);
	return readData();
}

int EntityDsp::readFragMips() {

	command(DSPP_HV_READ_FRAG_MIPS);
	return readData();
}

int EntityDsp::readDebugData() {

	command(DSPP_HV_READ_DEBUG_DATA);
	return readData();
}

int EntityDsp::readMaxInLevels() {

	command(DSPP_HV_READ_MAX_IN_LEVELS);
	return readData();
}

int EntityDsp::readMaxOutLevels() {

	command(DSPP_HV_READ_MAX_OUT_LEVELS);
	return readData();
}

int EntityDsp::readData() {

	int x = 0;
	dsp_read_data(dspHandle, &x, 1);
	
	return x;
}

void EntityDsp::voiceAssign(int VDIF, int STDF, int voice) { 

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - 
	// This function tells the dsp what voice is about to 
	// be referred to. The data sent now is VDIF, STDF and
	// the voice number. Subsequently, a voiceOn or voiceShift
	// command can be issued, containing pitch and velocity 
	// information. However, the DSP implicitely knows which
	// voice that velocity and pitch data is for, thanks to
	// the immediately preceding data.
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - 

	// - - - - - - - - - - - - - - - - - - - - - - - -
	// Transmit a 24 bit number containing three bytes
	// 
	// 	MS byte : mid byte : LS byte
	// 	  STDF  :   VDIF   :  voice    
	//
	// All three values are positive, 0..127 */
	// - - - - - - - - - - - - - - - - - - - - - - 

	int word24;

	word24  = voice;
	word24 |= VDIF << 8;
	word24 |= STDF << 16;

	update(DSPP_HV_VOICE_ASSIGN, word24); 
}

void EntityDsp::voiceOn(int pitch, int velocity) {
	
	// - - - - - - - - - - - - - - - - - - - - - - - 
	// Send a 24-bit number containing two numbers
	//
	// 	MS byte :      LS word      
	//	velocty :       pitch	
	//		    
	// The velocity is 0..127 placed in the MS byte 
	// Pitch is signed 16-bit cents, in the LS word 
	// - - - - - - - - - - - - - - - - - - - - - - - 

	int word24 = (velocity << 16) | (pitch & 0xffff);

	update(DSPP_HV_VOICE_ON, word24);
}

void EntityDsp::voiceOnSmooth(int pitch, int velocity) {

	// Like voiceOn, but envelopes are set to attack
	// phase without resetting their levels to zero
	
	int word24 = (velocity << 16) | (pitch & 0xffff);

	update(DSPP_HV_VOICE_ON_SMOOTH, word24);
}

void EntityDsp::voiceShift(int pitch) {
	
	// We don't need velocity when shifting a note
	
	update(DSPP_HV_VOICE_SHIFT, pitch);
}

void EntityDsp::voiceOff(int voice) { 
	
	update(DSPP_HV_VOICE_OFF, voice); 
}
	
void EntityDsp::voiceMute(int voice) { 
	
	update(DSPP_HV_VOICE_MUTE, voice); 
}

void EntityDsp::command(int hCommand) {

	if(!dsp_write_command(dspHandle, hCommand>>1, true))
		TRACE("ERROR: Could not call DSP host command: ");

	#ifdef DEBUG_DSP
	else	TRACE("OKAY: DSP host command successful: ");

	switch(hCommand) {

		case DSPP_HV_CTR1 :			TRACE("DSPP_HV_CTR1\n"); break;
		case DSPP_HV_CTR2 :			TRACE("DSPP_HV_CTR2\n"); break;
		case DSPP_HV_CTR3 :			TRACE("DSPP_HV_CTR3\n"); break;
		case DSPP_HV_MMOD :			TRACE("DSPP_HV_MMOD\n"); break;
		case DSPP_HV_NOTE_ON :			TRACE("DSPP_HV_NOTE_ON\n"); break;
		case DSPP_HV_NOTE_SHIFT :		TRACE("DSPP_HV_NOTE_SHIFT\n"); break;
		case DSPP_HV_NOTE_OFF :			TRACE("DSPP_HV_NOTE_OFF\n"); break;
		case DSPP_HV_NEW_SOUND_NOTE :		TRACE("DSPP_HV_NEW_SOUND_NOTE\n"); break;
		case DSPP_HV_TRANSFER_XDATA :		TRACE("DSPP_HV_TRANSFER_XDATA\n"); break;
		case DSPP_HV_TRANSFER_YDATA :		TRACE("DSPP_HV_TRANSFER_YDATA\n"); break;
		case DSPP_HV_TRANSFER_PDATA :		TRACE("DSPP_HV_TRANSFER_PDATA\n"); break;
		case DSPP_HV_MINIMISE_DSP :		TRACE("DSPP_HV_MINIMISE_DSP\n"); break;
		case DSPP_HV_MAXIMISE_DSP :		TRACE("DSPP_HV_MAXIMISE_DSP\n"); break;
		case DSPP_HV_SOUND_PENDING :		TRACE("DSPP_HV_SOUND_PENDING"); break;
		case DSPP_HV_READ_MAX_IN_LEVELS :	TRACE("DSPP_HV_READ_MAX_IN_LEVELS\n"); break;
		case DSPP_HV_READ_MAX_OUT_LEVELS :	TRACE("DSPP_HV_READ_MAX_OUT_LEVELS\n"); break;
		case DSPP_HV_PITCH_SHIFT :		TRACE("DSPP_HV_PITCH_SHIFT\n"); break;
		case DSPP_HV_MASTER_VOL :		TRACE("DSPP_HV_MASTER_VOL\n"); break;
		case DSPP_HV_MASTER_TUNE :		TRACE("DSPP_HV_MASTER_TUNE\n"); break;
		case DSPP_HV_READ_TOTAL_MIPS :		TRACE("DSPP_HV_READ_TOTAL_MIPS\n"); break;
		case DSPP_HV_READ_FRAG_MIPS :		TRACE("DSPP_HV_READ_FRAG_MIPS\n"); break;
		case DSPP_HV_READ_DEBUG_DATA :		TRACE("DSPP_HV_READ_DEBUG_DATA\n"); break;
	}
	
	#endif
}
