/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityEditMode.cpp							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005                                      *
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/

#include "EntityEditMode.h"

EntityEditMode::EntityEditMode(EntityPanel *panel_, EntityGlobal* global_, ParaStructure *paras_) : EntityMode(panel_) {

	global = global_;
	paras = paras_;

	firstEdit = true;
	matrixEdit = false;

	stackParameter = paras->getParaByName("stack");
	
	parameterIndex = 0;
	paras->setCurrentGroupPageParaByIndex(parameterIndex);
	
	group = paras->getCurrentGroup();
	page = group->getCurrentPage();
	parameter = page->getCurrentParameter();
	source = -1;

	levelMeter = new EntityLevelMeter;

	source = -1;
}

Mode EntityEditMode::enter(EntitySound *sound_) {

	sound = sound_;

	matrixUsageDisplay = false;
	
	panel->setLEDs(PANEL01_LED_EDIT);
	
	if(global->shiftKey()) panel->setLEDs(PANEL01_LED_SHIFT);
	
	setPotsByContext();

	if(firstEdit) { panel->requestPotUpdates(); firstEdit = false; }
	
	checkLevelMeter();
	
	processEvents(); 	

	panel->transmitPot3EventsAs(PANEL01_POT_CTRL3); 
	panel->clearLEDs(PANEL01_LED_SHIFT);
	panel->setPotsForModulation();

	return rMode;
}

void EntityEditMode::setPotsByContext() {

	if(!global->fastEdit()) {
		
		panel->setPotsForModulation();
		return;
	}
	
	panel->setPotsForData(PANEL01_POT_CTRL1, PANEL01_POT_CTRL2);
	
	if(matrixEdit) { 
		
		panel->setPotsForData(PANEL01_POT_VOLUME);

		switch(source) {		

			case CTR1 	: panel->transmitPot3EventsAs(PANEL01_POT_CTRL1); break;
			case CTR2 	: panel->transmitPot3EventsAs(PANEL01_POT_CTRL2); break;
			default 	: panel->transmitPot3EventsAs(PANEL01_POT_CTRL3); break;
		}
	}
	
	else {

		panel->setPotsForModulation(PANEL01_POT_VOLUME, PANEL01_POT_CTRL3);
		panel->transmitPot3EventsAs(PANEL01_POT_CTRL3); 
	}
}

void EntityEditMode::displayUpdate() {

	/* Avoid matrix edit for non dsp parameters */
	
	if(matrixEdit) {

		if(!page->getCurrentParameter()->isDspParameter()) {

			/* Use old index to return to previous parameter */
			paras->setCurrentGroupPageParaByIndex(parameterIndex);

			/* No need to update display */
			return;
		}
	}
	
	/* Update parameter */

	group = paras->getCurrentGroup();
	page = group->getCurrentPage();
	parameter = page->getCurrentParameter();
	source = matrixEdit ? paras->getCurrentSource() : -1;

	parameterIndex = parameter->getParaIndex();
	
	if(matrixEdit) panel->setEncoderAcceleration(100);		
	else panel->setEncoderAcceleration(parameter->getAcceleration());		
	
	setPotsByContext();	// Possibly swap active edit pots
	
	/* Initiate or suspend level metering */

	bool meterWasActive = levelMeter->isActive();
	checkLevelMeter();
	
	/* Update synth with selected parameter and source */

	sm[0] = PARAMETER_SELECT;
	sm[1] = parameter->getDspParaIndex();
	sm[2] = source;
	
	sendMessage(synq, sm, 3);

	/* Update LCD */
	
	char 	line[PANEL_LCD_MAX_LINE_LEN + 7];
	
	char*	s;	// string to copy
	int 	p;	// Copy position
	int 	e;	// end of space
	int 	vl; 	// value length
		
	if(meterWasActive && levelMeter->isActive()) panel->clearLine(1);
	else panel->clear();
	
	if(!matrixEdit) { 
		
		/* Top line */
	
		s = page->getName();
		p = 0; while(*s) line[p++] = *s++;	// Copy str but not \0

		s = page->getShortName();
		e = PANEL_LCD_MAX_LINE_LEN - strlen(s);
		while(p < e) line[p++] = 32;		// Spaces to shortname
		while((line[p++] = *s++));		// Copy str inc \0;
			
		if(!levelMeter->isActive()) panel->print(0, 0, line);

		/* Bottom line */
		
		s = parameter->getName(3);
		p = 0; while(*s) line[p++] = *s++;	// Copy str but not \0

		s = sound->getValStr(parameter, source, vl);
		e = PANEL_LCD_MAX_LINE_LEN - vl;

		while(p < e) line[p++] = 32;

		while(*s) line[p++] = *s++;	
		line[p++] = 0;
		
		panel->print(1, 0, line);
		
	} else { 

		/* Top line */
		
		s = paras->getSourceName(source);
		p = 0; while(*s) line[p++] = *s++;

		line[p++] = 32;
		line[p++] = 1;	// matrix route symbol
		line[p++] = 32;
		
		s = page->getShortName();
		if(*s) {
			while(*s) line[p++] = *s++;
			line[p++] = 32;
		}
		
		s = parameter->getShortName();
		while(*s) line[p++] = *s++;
		line[p++] = 0;
			
		panel->print(0, 0, line);

		/* Bottom line */

		s = sound->getValStr(parameter, source, vl);
		e = PANEL_LCD_MAX_LINE_LEN - vl - 7;

		p = 0; while(p < e) line[p++] = 32;

		while(*s) line[p++] = *s++;		
		line[p++] = 0;
		panel->print(1, 0, line);
	}
}

void EntityEditMode::displayValueUpdate() {

	/* Tell synth to transfer parameter value to dsp */

	sendMessage(synq, PARAMETER_MODIFY);

	/* Update display */
	
	if(matrixUsageDisplay) {
		displayUpdate();
		matrixUsageDisplay = FALSE;
		return;
	}

	int vl, p;
	char *s = sound->getValStr(parameter, source, vl);
	p = PANEL_LCD_MAX_LINE_LEN - vl;
	if(matrixEdit) p -= 7;
	panel->print(1, p, s);
}	

void EntityEditMode::displayMatrixUsage() {

	char 	line[PANEL_LCD_MAX_LINE_LEN + 7];
	
	int n = sound->getnMatrixLinks();
	int p = (10000 * n) / MAX_MATRIX_LINKS;

	panel->clear();
	
	sprintf(line, "  Matrix %i.%2.2i%%   ", p / 100, p % 100);
	panel->print(0, 0, line);

	sprintf(line, "   (%i links)    ", n);
	panel->print(1, 0, line);

	matrixUsageDisplay = true;
	levelMeter->deActivate();
}

Status EntityEditMode::incrementValue(int i) {

	return sound->incValue(parameter, i, source);	
}

void EntityEditMode::edit() { 

	if(matrixEdit) {
		
		matrixEdit = false;
		source = -1;
		displayUpdate();								

		return;
	}
		
	rMode = SOUND; 
}

void EntityEditMode::shift() {

	if(global->toggleShiftKey()) panel->setLEDs(PANEL01_LED_SHIFT);
	else panel->clearLEDs(PANEL01_LED_SHIFT);
}

void EntityEditMode::shiftPartUp() {

	source = paras->getCurrentSource();
	
	if(matrixEdit) sound->findNextMatrixDestination(source, parameter);
	else if(!sound->getValue(parameter, source)) sound->findNextMatrixDestination(source, parameter);
	
	matrixEdit = true;
	displayUpdate();
														
	panel->repeat(AUTO_MATRIX_UP, 825, 210);
}

void EntityEditMode::shiftPartDown() {

 	source = paras->getCurrentSource();

	if(matrixEdit) source = sound->findNextMatrixSource(source, parameter);
	else if(!sound->getValue(parameter, source)) sound->findNextMatrixSource(paras->getCurrentSource(), parameter);

	matrixEdit = true;
	displayUpdate();

	panel->repeat(AUTO_MATRIX_DOWN, 825, 210);
}

void EntityEditMode::partUp() { 

	if(!parameter->isDspParameter()) {
	
		if(parameter == stackParameter) rMode = STACK;
		return;
	}
			
	if(global->shiftKey()) { shiftPartUp(); return; }
	
	if(keyStates & PANEL01_KEY_PART_DOWN) {
								
		displayMatrixUsage();
		return;
	}
	
	if(!matrixEdit) {

 		source = paras->getCurrentSource();
		matrixEdit = true;
		displayUpdate();
		return;
	}
	
	sourceInc(+1); 
}

void EntityEditMode::partDown() { 
	
	if(!parameter->isDspParameter()) {
	
		if(parameter == stackParameter) rMode = STACK;
		return;
	}
			
	if(global->shiftKey()) { shiftPartDown(); return; }
	
	if(keyStates & PANEL01_KEY_PART_UP) {
								
		displayMatrixUsage();
		return;
	}
	
	if(!matrixEdit) {

 		source = paras->getCurrentSource();
		matrixEdit = true;
		displayUpdate();
		return;
	}
	
	sourceInc(-1); 
}

void EntityEditMode::sourceInc(int x) {

	if(matrixEdit) if(paras->sourceInc(x)) {
			
		source = paras->getCurrentSource();
		displayUpdate();
	
		if(x == +1) panel->repeat(AUTO_SOURCE_UP, SLOW_REPEAT_TIME); 
		if(x == -1) panel->repeat(AUTO_SOURCE_DOWN, SLOW_REPEAT_TIME); 
	}
}

void EntityEditMode::groupUp() { 
	
	if(paras->groupInc(+1)) { 
		
		checkNextDspPara(+1);
		panel->repeat(AUTO_GROUP_UP, SLOW_REPEAT_TIME); 
		displayUpdate(); 
	} 
}

void EntityEditMode::shiftGroupUp() { 
	
	paras->setCurrentGroupByIndex(paras->getnGroups() - 1); 
	checkNextDspPara(+1);
	displayUpdate(); 
}

void EntityEditMode::groupUpKeyUp() { }

void EntityEditMode::groupDown() { 
	
	if(paras->groupInc(-1)) { 
		checkNextDspPara(-1);
		panel->repeat(AUTO_GROUP_DOWN, SLOW_REPEAT_TIME); 
		displayUpdate(); 
	} 
}

void EntityEditMode::shiftGroupDown() { 
	
	paras->setCurrentGroupByIndex(0); 
	checkNextDspPara(-1);
	displayUpdate(); 
}

void EntityEditMode::groupDownKeyUp() { }

void EntityEditMode::pageUp() { 
	
	if(group->pageInc(+1)) { 
		checkNextDspPara(+1);
		panel->repeat(AUTO_PAGE_UP, SLOW_REPEAT_TIME); 
		displayUpdate(); 
	} 
}

void EntityEditMode::shiftPageUp() { 
	
	group->setCurrentPageByIndex(group->getnPages() - 1); 
	checkNextDspPara(+1);
	displayUpdate(); 
}

void EntityEditMode::pageDown() { 
	
	if(group->pageInc(-1)) { 
		checkNextDspPara(-1);
		panel->repeat(AUTO_PAGE_DOWN, SLOW_REPEAT_TIME); 
		displayUpdate(); 
	} 
}

void EntityEditMode::shiftPageDown() { 
	
	group->setCurrentPageByIndex(0); 
	checkNextDspPara(-1);
	displayUpdate(); 
}

void EntityEditMode::paramUp() { 
	
	if(page->parameterInc(+1)) { 
		checkNextDspPara(+1);
		panel->repeat(AUTO_PARAM_UP, SLOW_REPEAT_TIME); 
		displayUpdate(); 
	} 
}

void EntityEditMode::shiftParamUp() { 
	
	page->setCurrentParameterByIndex(page->getnParas() - 1); 
	checkNextDspPara(+1);
	displayUpdate(); 
}

void EntityEditMode::paramDown() { 
	
	if(page->parameterInc(-1)) { 
		checkNextDspPara(-1);
		panel->repeat(AUTO_PARAM_DOWN, SLOW_REPEAT_TIME); 
		displayUpdate(); 
	} 
}

void EntityEditMode::shiftParamDown() { 
	
	page->setCurrentParameterByIndex(0); 
	checkNextDspPara(-1);
	displayUpdate(); 
}

void EntityEditMode::checkNextDspPara(int direction) { 

	/* This is a helper function - when selecting group, page or parameter,
	 * it is possible to hit a non-dsp parameter. If in matrix edit, the 
	 * non-dsp parameter needs to be skipped over, to the next dsp parameter. */

	if(!matrixEdit) return;

	Group *gp = paras->getCurrentGroup();
	Page *pg = gp->getCurrentPage();
	
	/* Non-dsp paras are guarded below and above by dsp paras, so a search in either
	 * direction will always yield a dsp parameter eventually (the non-dsp paras are 
	 * "Polyphony" and "Stack") */
	
	while(!pg->getCurrentParameter()->isDspParameter()) pg->parameterInc(direction);
}

void EntityEditMode::valueUp() {

	if(!global->shiftKey()) { 
		
		panel->repeat(AUTO_VALUE_UP, 825, 50);
		incrementEvent();
	}

	else unMute();
}

void EntityEditMode::valueDown() {

	if(!global->shiftKey()) {	
	
		panel->repeat(AUTO_VALUE_DOWN, 825, 50);
		decrementEvent();
	}

	else mute();
}

void EntityEditMode::shiftValueUp() {

	unMute();
}

void EntityEditMode::shiftValueDown() {

	mute();
}

void EntityEditMode::unMute() {

	if(sound->getMute(parameter, source)) {

		if(sound->setMute(parameter, FALSE, source) == PARAMETER_EXCEPTION) {

			parameterException();
			return;
		}

		displayValueUpdate();
		return;
	}
}

void EntityEditMode::mute() {

	if(!sound->getMute(parameter, source))
		if(sound->setMute(parameter, TRUE, source))
			displayValueUpdate();
}
	
void EntityEditMode::pot0Event(int p) {
	
	if(matrixEdit) {
		
		if(paras->setCurrentSourceByPot(p)) {
			
			source = paras->getCurrentSource();
			displayUpdate();
		}
	}
}

void EntityEditMode::pot1Event(int p) { if(page->setCurrentParameterByPot(p)) displayUpdate(); } 

void EntityEditMode::pot2Event(int p) { if(paras->setCurrentPageByPot(p)) displayUpdate(); }

void EntityEditMode::editShift() {
	
	if(!global->toggleFastEdit()) panel->setPotsForModulation();
	else setPotsByContext();
}

void EntityEditMode::parameterException() {

	/* Clear any button autorepeats */
	panel->clearRepeat();

	/* Warn user */
	panel->print(1, 0 ,"  MATRIX FULL");

	/* Delay before restoring display */
	panel->delay(DISPLAY_UPDATE, 725);
}

void EntityEditMode::buttonOffEvent() {

	if(matrixUsageDisplay) {
						
		displayUpdate();
		matrixUsageDisplay = false;
	}
}

void EntityEditMode::checkLevelMeter() {

	/* Is a metered parameter (ie Volume)? */

	if(source == -1 && parameter->isMeteredParameter()) {

		levelMeter->activate();
		sendMessage(synq, CLEAR_DSP_LEVELS);
		panel->request(READ_OUTPUT_LEVELS, 25, 100);
	}

	else {
		levelMeter->deActivate();
		panel->clearRequest();
	}	
}

void EntityEditMode::autoEvent(int v) {

	switch(v) {

		case DISPLAY_MIDI_EDIT: 	displayMidiEdit(); break;

		case AUTO_OUTPUT_LEVELS:	if(levelMeter->isActive()) {
						
							panel->print(0, 0, levelMeter->draw(mBuffer[1], mBuffer[2]));
						}
					
						break;
					
		case AUTO_VALUE_UP:		incrementEvent(); break;
						
		case AUTO_VALUE_DOWN:		decrementEvent(); break;

		case AUTO_MATRIX_UP:		source = paras->getCurrentSource();
						sound->findNextMatrixDestination(source, parameter);
						displayUpdate();
						break;

		case AUTO_MATRIX_DOWN:		sound->findNextMatrixSource(paras->getCurrentSource(), parameter);
						displayUpdate();
						break;

		case DISPLAY_UPDATE:		panel->clearDelay(); displayUpdate(); break;

		case AUTO_SOURCE_UP:		sourceInc(+1); break;
		case AUTO_SOURCE_DOWN:		sourceInc(-1); break;
		case AUTO_GROUP_UP:		groupUp(); break;
		case AUTO_GROUP_DOWN:		groupDown(); break;
		case AUTO_PAGE_UP:		pageUp(); break;
		case AUTO_PAGE_DOWN:		pageDown(); break;
		case AUTO_PARAM_UP:		paramUp(); break;
		case AUTO_PARAM_DOWN:		paramDown(); break;

	}
}

void EntityEditMode::displayMidiEdit() {

	/* The edited parameter and the midi edit parameter are the same, so the 
	 * the display needs updating. The dsp is updated independently. */

	if(matrixUsageDisplay) { displayUpdate(); matrixUsageDisplay = FALSE; return; }

	int vl, p;
	char *s = sound->getValStr(parameter, source, vl);
	p = PANEL_LCD_MAX_LINE_LEN - vl;
	if(matrixEdit) p -= 7;
	panel->print(1, p, s);
}	

