/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityFlash.cpp							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/
  
#include "EntityFlash.h"

char EntityString[] = "ENTITY";

EntityFlash::EntityFlash() {

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
	// Entity Flash File structure (version 1)
	//
	//		   Size in bytes			     Byte offset

	header.idString.size	= 6;	header.idString.offset		= 0;
	header.version.size 	= 1;	header.version.offset 		= 6;
	header.type.size 	= 1;	header.type.offset		= 7;
	header.index.size	= 2;	header.index.offset		= 8;
	header.length.size	= 2;	header.length.offset		= 10;
	header.transaction.size	= 4;	header.transaction.offset	= 12;

	header.size 		= 16; 

	// idString:	ENTITY (6 chars)
	// version: 	Entity model number (see EntityVersion.h)
	// type:	1 = global, 2 = sound
	// index:	specifies sound number, ie A10 = 9, B2 = 43
	// length:	length of file, including header and checksum
	// transaction:	32 bit transaction number
	//
	// All multibyte header data is stored with the most significant 
	// byte in the lower address, or earlier, if streamed.
	//
	// After the header there follows data of the length specified in 
	// the header, including a CRC32 checksum in the final four bytes.

	checksumSize 		= 4;

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	handle = flash_init();	
	
	unsigned int flashSize;
	flash_get_size(handle, &flashSize);

	int soundFileSpace = FLASH_MAX_SOUNDS * ENTITY_FLASH_FILE_SIZE;
	int spareFiles = (ENTITY_FLASH_MEMORY_SIZE - soundFileSpace) / ENTITY_FLASH_FILE_SIZE;
	int possibleFiles = flashSize / ENTITY_FLASH_FILE_SIZE;
	int versionFiles = 70; // Preferred minimum no. of versions, inc one global file.
	int banks = FLASH_MAX_SOUNDS / SOUND_BANK_SIZE;

	if(flashSize < ENTITY_FLASH_MEMORY_SIZE) TRACE("ERROR: %i bytes requested\n", ENTITY_FLASH_MEMORY_SIZE);
	if(spareFiles < versionFiles) TRACE("WARNING: only %i spare file locations\n", spareFiles);
	
	TRACE("Flash memory available = %i bytes\n", flashSize);
	TRACE("Could contain %i files\n", possibleFiles);
	TRACE("Flash file system using %i bytes ", ENTITY_FLASH_MEMORY_SIZE);
	TRACE("arranged as %i sounds, in %i banks of %i\n", FLASH_MAX_SOUNDS, banks, SOUND_BANK_SIZE);
	TRACE("with room for %i stored versions + one global\n", spareFiles - 1);

	presets = new EntityPresets;	// Library of non-writable sounds
	
	scan();
}

void EntityFlash::scan() {
	
	unsigned int maxTrans[FLASH_MAX_SOUNDS]; 
	
	HeaderInfo file;
	
	nextTransactionNumber = 1;
	nVersions = 0;
	nSpare = 0; 
	
	for(int x = 0; x < FLASH_MAX_SOUNDS; x++) {
		sound[x] = -1;
		maxTrans[x] = 0;
	}
	
	unsigned int maxTransGlobal = 0;
	globalLocation = -1;
	
	for(int x = 0; x < FLASH_N_FILES; x++) {
	
		if(!getHdrInfo(x, file)) { nSpare++; spare[x] = true; continue; } 

		if((spare[x] = !verify(x))) { nSpare++; continue; }	

		if(file.transaction >= nextTransactionNumber) nextTransactionNumber = file.transaction + 1;

		if(file.type == FLASH_GLOBAL_FILE) {

			if(file.transaction > maxTransGlobal) {	 	// greatest trans no. for global file so far?
				maxTransGlobal = file.transaction; 	// if so, update transaction no.
				globalLocation = x;			// and record location as current for global
			}
		}

		else if(file.type == FLASH_SOUND_FILE) {

			if(file.transaction > maxTrans[file.index]) {	 // greatest trans no. for this sound so far?
				maxTrans[file.index] = file.transaction; // if so, update transaction no.
				sound[file.index] = x;			 // and record location as most recent for this sound
			}
			
			#ifdef DEBUG_FLASH

			flash_read_data(handle, x * ENTITY_FLASH_FILE_SIZE + header.size, buf, SOUND_NAME_LENGTH);
			TRACE("%i found at location %i, name = '%s'\n", file.index, x, (char*) buf);
			
			#endif
		}
	}
}

void EntityFlash::setPriority(rtems_task_priority new_priority) {

	rtems_id 		tfla;
	rtems_status_code 	status;
	rtems_task_priority	old_priority;
       
	/* This function solves a task prioritisation problem. When a flash handle is 
	 * obtained, a dormant task is created by the code behind the Soundart flash API. 
	 * When writing to flash, flash_write_data() returns immediately, having handed 
	 * over the writing operation (which can last several seconds) to this task. 
	 * This task is called TFLA, and is initialised with the highest RTEMS priority 
	 * level of 1. As a result, flash operations can glitch the UI and other tasks, 
	 * or, when dependencies are complex, freeze the system (though this is a bug). 
	 * The solution is to modify the priority level of TFLA after initialisation, 
	 * using this function. It is called from EntityControl.cpp */ 

	status = rtems_task_ident(rtems_build_name('T','F','L','A'), RTEMS_SEARCH_ALL_NODES, &tfla);
	if(status != RTEMS_SUCCESSFUL) TRACE("ERROR: Cannot get task id for TFLA\n");

	rtems_task_set_priority(tfla, new_priority, &old_priority);
}
 
int EntityFlash::getnSounds() {

	return FLASH_MAX_SOUNDS + presets->getnPresets();
}

int EntityFlash::getnWritableSounds() {

	return FLASH_MAX_SOUNDS;
}

char *EntityFlash::snoopName(int soundNumber) {

	static rtems_unsigned8 nameBuf[SOUND_NAME_LENGTH];

	version = 0;
	nVersions = 0;
	
	if(soundNumber >= FLASH_MAX_SOUNDS) {
		
		return presets->snoopName(soundNumber - FLASH_MAX_SOUNDS);	
	}
	
	/* If sound file not present, use default name */
	if((location = sound[soundNumber]) < 0) return EntitySound::getDefaultName();

	#ifdef DEBUG_FLASH
	TRACE("File location = %i\n", location); 
	#endif

	flash_read_data(handle, location * ENTITY_FLASH_FILE_SIZE + header.size, nameBuf, SOUND_NAME_LENGTH);
	
	return (char*) nameBuf;
}

bool EntityFlash::load(EntityGlobal *global) {

	if(globalLocation < 0) { 
	
		TRACE("Loading default global values\n");	
		global->loadDefaultValues(); 
		return true; 
	}
	
	TRACE("Loading global flash file\n");

	int length = getFileLength(globalLocation) - header.size;
	
	flash_read_data(handle, globalLocation * ENTITY_FLASH_FILE_SIZE + header.size, buffer, length);

	global->unpack(buffer);

	return true;
}

int EntityFlash::load(int sn, rtems_unsigned8 *data) {

	/* This function is used when saving stored sounds to midi */

	int mLocation;
	
	if(sn >= FLASH_MAX_SOUNDS) return presets->load(data, sn - FLASH_MAX_SOUNDS);

	if((mLocation = sound[sn]) < 0) return presets->loadDefault(data);

	int length = getFileLength(mLocation) - header.size - checksumSize;
	
	flash_read_data(handle, mLocation * ENTITY_FLASH_FILE_SIZE + header.size, data, length);

	return length;
}

bool EntityFlash::load(EntitySound *snd, int sn) {

	if((soundNumber = sn) >= FLASH_MAX_SOUNDS) {
		
		return presets->load(snd, sn, sn - FLASH_MAX_SOUNDS);
	}

	version = 0;
	nVersions = 0;
	
	if((location = sound[soundNumber]) < 0) {

		/* File not in flash, so use default. There are two possible default 
		 * sources one is from a method in the sound object, and the other 
		 * is from EntityPresetData.h */   
		
		if(presets->loadDefault(snd, sn)) return true;		// Use default in preset data
		snd->loadDefaultValues(soundNumber); return true;	// else generate values
	}

	int length = getFileLength(location) - header.size;
	
	flash_read_data(handle, location * ENTITY_FLASH_FILE_SIZE + header.size, buffer, length);

	snd->deCompress(buffer);
	snd->setNumber(soundNumber);

	return true;
}

int EntityFlash::initVersionAccess(int n) { 
	
	static QuickSort<int, unsigned int> qs(versionLocation, versionTransaction);
	HeaderInfo file;

	if((soundNumber = n) >= FLASH_MAX_SOUNDS) return false;

	if((location = sound[soundNumber]) < 0) {

		/* File not in flash, so indicate use of default version */

		versionLocation[nVersions = 1] = -1;
		return true;
	}

	/* Locate all versions of loaded sound */
	
	nVersions = 0;
	
	for(int x = 0; x < FLASH_N_FILES; x++) {
			
		flash_read_data(handle, x * ENTITY_FLASH_FILE_SIZE + header.index.offset, buf, header.index.size);
			
		unsigned int s = header.index.size, o = 0, v = 0;
		while(s--) v = buf[o++] | (v << 8);

		if(v == (unsigned int)soundNumber) {

			if(!getHdrInfo(x, file)) continue; 
			if(file.type != FLASH_SOUND_FILE) continue;

			versionLocation[nVersions] = x;
			versionTransaction[nVersions] = file.transaction;

			nVersions++;
		}
	}

	if(!nVersions) TRACE("BUG in EntityFlash::initVersionAccess()\n");

	/* Sort versions to put most recent at top of list */
	if(nVersions > 1) qs.sort(nVersions); 

	/* Limit number of versions because of LCD size */
	nVersions = nVersions > 99 ? 99 : nVersions;
	
	return nVersions;	
}

bool EntityFlash::loadNextVersion(EntitySound *snd) {

	if(version <= 0) return false;
	return loadVersion(snd, --version);
}

bool EntityFlash::loadPreviousVersion(EntitySound *snd) {

	if(version >= (nVersions-1)) return false;
	return loadVersion(snd, ++version);
}

bool EntityFlash::loadVersion(EntitySound *snd, int v) {

	if(!nVersions) return false;
	if(v < 0) return false;
	if(v >= nVersions) return false;
	
	version = v;
	
	if((location = versionLocation[v]) < 0) {
		
		/* File not in flash, so use default */
		snd->loadDefaultValues(soundNumber);
		return true;
	}

	int length = getFileLength(location) - header.size;

	if(!length) return false;
	
	flash_read_data(handle, location * ENTITY_FLASH_FILE_SIZE + header.size, buffer, length);

	snd->deCompress(buffer);
	snd->setNumber(soundNumber);

	return true;
}	

bool EntityFlash::save(rtems_unsigned8 *data, int nBytesData) {

	// This function is used for saving global settings loaded from midi
	
	HeaderInfo 	file;
	
	if(nBytesData > (ENTITY_FLASH_FILE_SIZE - header.size - checksumSize)) {
	
		TRACE("ERROR: too much global data (%i bytes)\n", nBytesData);
	}

	for(int x = 0, y = header.size; x < nBytesData;) buffer[y++] = data[x++];
	
	file.idString	= EntityString;
	file.version 	= ENTITY_MODEL_NUMBER;
	file.type	= FLASH_GLOBAL_FILE;
	file.index 	= 0;
	file.length	= header.size + nBytesData + checksumSize;
	file.transaction = nextTransactionNumber++;

	if(nextTransactionNumber >= FLASH_MAX_TRANSACTIONS) {

		fixTransactionNumbers();
		file.transaction = nextTransactionNumber++;
	}
	
	globalLocation = newLocation(); if(!writeFile(buffer, file, globalLocation)) return false;
	
	return true;
}

bool EntityFlash::save(EntityGlobal *global) {

	HeaderInfo 	file;
	
	int nBytesData = global->pack(&buffer[header.size]);
	
	if(nBytesData > (ENTITY_FLASH_FILE_SIZE - header.size - checksumSize)) {
	
		TRACE("ERROR: too much global data (%i bytes)\n", nBytesData);
	}

	file.idString	= EntityString;
	file.version 	= ENTITY_MODEL_NUMBER;
	file.type	= FLASH_GLOBAL_FILE;
	file.index 	= 0;
	file.length	= header.size + nBytesData + checksumSize;
	file.transaction = nextTransactionNumber++;

	if(nextTransactionNumber >= FLASH_MAX_TRANSACTIONS) {

		fixTransactionNumbers();
		file.transaction = nextTransactionNumber++;
	}
	
	globalLocation = newLocation(); if(!writeFile(buffer, file, globalLocation)) return false;
	
	return true;
}

bool EntityFlash::save(EntitySound *snd) {

	HeaderInfo 	file;
	
	if(snd->getNumber() >= FLASH_MAX_SOUNDS) return false;

	int nBytesData = snd->compress(&buffer[header.size]);
	
	if(nBytesData > (ENTITY_FLASH_FILE_SIZE - header.size - checksumSize)) {
	
		TRACE("ERROR: too much compressed sound data (%i bytes)\n", nBytesData);
		return false;
	}

	file.idString 	= EntityString;
	file.version 	= ENTITY_MODEL_NUMBER;
	file.type	= FLASH_SOUND_FILE;
	file.index 	= snd->getNumber();
	file.length	= header.size + nBytesData + checksumSize;
	file.transaction = nextTransactionNumber++;

	if(nextTransactionNumber >= FLASH_MAX_TRANSACTIONS) {

		fixTransactionNumbers();
		file.transaction = nextTransactionNumber++;
	}
	
	location = newLocation(); if(!writeFile(buffer, file, location)) return false;
	sound[snd->getNumber()] = location; // Update location of current version of this sound

	version = 0;
	nVersions = 0;
	
	return true;
}

bool EntityFlash::save(int soundNumber, rtems_unsigned8 *data, int nBytesData) {

	// This function is used for saving sounds loaded from midi
 
	HeaderInfo 	file;
	
	if(soundNumber >= FLASH_MAX_SOUNDS) return false;

	if(nBytesData > (ENTITY_FLASH_FILE_SIZE - header.size - checksumSize)) {
	
		TRACE("ERROR: too much sound data (%i bytes)\n", nBytesData);
		return false;
	}

	for(int x = 0, y = header.size; x < nBytesData;) buffer[y++] = data[x++];
	
	file.idString 	= EntityString;
	file.version 	= ENTITY_MODEL_NUMBER;
	file.type	= FLASH_SOUND_FILE;
	file.index 	= soundNumber;
	file.length	= header.size + nBytesData + checksumSize;
	file.transaction = nextTransactionNumber++;

	if(nextTransactionNumber >= FLASH_MAX_TRANSACTIONS) {

		fixTransactionNumbers();
		file.transaction = nextTransactionNumber++;
	}

	int mLocation = newLocation(); if(!writeFile(buffer, file, mLocation)) return false;
	sound[soundNumber] = mLocation; // Update location of current version of this sound

	version = 0;
	nVersions = 0;
	
	return true;
}

void EntityFlash::fixTransactionNumbers() {

	TRACE("Re-writing transaction numbers\n");
	TRACE("BUG: EntityFlash::fixTransactionNumbers() not written yet\n");
	
	// Well, this hasn't been written yet, and perhaps never will.
	// To do it, load all data flash contents into RAM, sort
	// files into order looking only at their transaction numbers,
	// rewrite their transaction numbers starting at 1, then write
	// the entire contents back to flash in one go. There is no
	// need to change file locations, only transaction numbers.

	// With a thirty-two bit transaction number, this shouldn't 
	// be needed within the lifetime of the flash chip.
}

bool EntityFlash::writeFile(unsigned char *filebuf, HeaderInfo &file, int fLocation) {
	
	unsigned int s;		// size of field
	unsigned int o;		// offset of field

	int k; 			// general purpose counter
	
	// Add header info to start of buffer

	s = header.idString.size; o = header.idString.offset; k = 0; 
	while(s--) filebuf[o++] = file.idString[k++];
	
	s = header.version.size; o = header.version.offset;
	while(s--) filebuf[o++] = (file.version >> (s << 3));
			
	s = header.type.size; o = header.type.offset;
	while(s--) filebuf[o++] = (file.type >> (s << 3));
			
	s = header.index.size; o = header.index.offset;
	while(s--) filebuf[o++] = (file.index >> (s << 3));

	s = header.length.size; o = header.length.offset;
	while(s--) filebuf[o++] = (file.length >> (s << 3));
			
	s = header.transaction.size; o = header.transaction.offset;
	while(s--) filebuf[o++] = (file.transaction >> (s << 3));
			
	// Add checksum

	unsigned int c = crc.checksum(filebuf, file.length - checksumSize);
	
	s = checksumSize; o = file.length - checksumSize;
	while(s--) filebuf[o++] = (c >> (s << 3));

	// Identify start address in flash memory where file will be written
	
	if(spare[fLocation]) { nSpare--; spare[fLocation] = false; }
	unsigned int addr = fLocation * ENTITY_FLASH_FILE_SIZE;
	
	// Write to flash
	
	flash_write_data(handle, (rtems_unsigned32) addr, (rtems_unsigned8*) filebuf, (rtems_unsigned32) file.length);
	
	return true;	
}

int EntityFlash::newLocation() {

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// This is the simplest (and best?) strategy for overwriting old files:
	//
	// 1. If there are any spare locations in flash, use them first.
	// 2. If there is more than one global file present, overwrite the oldest version. 
	// 3. Identify oldest of any out-of-date sound files.
	
	HeaderInfo file;
	int fLocation;

	int		oldestVersion[FLASH_MAX_SOUNDS];	
	unsigned int 	minTransSound[FLASH_MAX_SOUNDS];
	
	for(int x = 0; x < FLASH_MAX_SOUNDS; x++) {
		oldestVersion[x] = -1;
		minTransSound[x] = FLASH_MAX_TRANSACTIONS;
	}
	
	int		nGlobal = 0;
	int		oldestGlobal = -1;
	unsigned int	minTransGlobal = FLASH_MAX_TRANSACTIONS;

	for(fLocation = 0; fLocation < FLASH_N_FILES; fLocation++) {
	
		// 1. If flash location is unwritten or unformatted
		// then the destination has been found

		if(spare[fLocation]) return fLocation;
		if(!getHdrInfo(fLocation, file)) return fLocation;
		
		// 2. There is only one global file, and any old versions
		// can be immediately overwritten, if available

		if(file.type == FLASH_GLOBAL_FILE) {

			if(file.transaction < minTransGlobal) {
				if(++nGlobal > 1) return fLocation;
				minTransGlobal = file.transaction;
				oldestGlobal = fLocation;
				continue;
			}
		}
		
		// 3(a). If this is not the most recent version of a sound, it is
		// a candidate for overwriting, so check if it is oldest so far

		if(fLocation != sound[file.index]) { // if not most recent version
			if(file.transaction < minTransSound[file.index]) {
				minTransSound[file.index] = file.transaction;
				oldestVersion[file.index] = fLocation;
				continue;
			}
		}
	}
			
	// 3(b). Look in list of oldest of out-of-date
	// version found in 3(a) and select oldest
		
	unsigned int minTrans = FLASH_MAX_TRANSACTIONS;
	for(int x = 0; x < FLASH_MAX_SOUNDS; x++) {
		if(minTransSound[x] < minTrans) {
			minTrans = minTransSound[x];
			fLocation = oldestVersion[x];
		}
	}

	return fLocation;	
}

bool EntityFlash::verify(int fLocation) {
	
	if(fLocation >= FLASH_N_FILES) return 0;	

	int length = getFileLength(fLocation);

	if((length < 0) || (length > ENTITY_FLASH_FILE_SIZE)) return false;
	
	int offset = fLocation * ENTITY_FLASH_FILE_SIZE;
	
	if(!flash_read_data(handle, offset, buffer, length)) {

		TRACE("Could not read specified flash memory from %i to %i\n", offset, length + offset);
	}
	
	length -= checksumSize; int i = length; // exclude checksum at end of file
	
	int s = checksumSize; 
	unsigned int checksum = 0;
	while(s--) checksum = buffer[i++] | (checksum << 8);	// MSB is at addr, NSB is at addr+1...
	
	if(crc.checksum(buffer, length) != checksum) return false;
	
	return true;
}

int EntityFlash::validSoundNumber(int x) {

	int max = FLASH_MAX_SOUNDS + presets->getnPresets() - 1; 
	
	if(x < 0) return 0;
	if(x > max) return max;
	
	return x;
}

int EntityFlash::validSaveSoundNumber(int x) {

	int max = FLASH_MAX_SOUNDS - 1; 
	
	if(x < 0) return 0;
	if(x > max) return max;
	
	return x;
}

int EntityFlash::getFileLength(int fLocation) {

	int offset = fLocation * ENTITY_FLASH_FILE_SIZE + header.length.offset;
	flash_read_data(handle, offset, buf, header.length.size);

	int s = header.length.size; int o = 0; int v = 0;
	while(s--) v = buf[o++] | (v << 8);	// MSB is at addr, NSB is at addr+1...
	
	if(v > ENTITY_FLASH_FILE_SIZE) {
	       	
		TRACE("ERROR: EntityFlash::getFileLength return value too large for fs\n");
		v = 0; // At least try to minimise the damage (we should never get here)
	}

	return v;
}

bool EntityFlash::getHdrInfo(int fLocation, HeaderInfo &h) { 

	unsigned int o;		// field offset
	unsigned int s;		// size of field
	unsigned int v;		// value

	int k; 			// general purpose counter
	
	int offset = fLocation * ENTITY_FLASH_FILE_SIZE;

	flash_read_data(handle, offset, buf, header.size);

	s = header.idString.size; o = header.idString.offset; k = 0; 
	while(s--) if(buf[o++] != EntityString[k++]) return false;
	
	s = header.version.size; o = header.version.offset; v = 0;
	while(s--) v = buf[o++] | (v << 8);
	h.version = v;
		
	s = header.type.size; o = header.type.offset; v = 0;
	while(s--) v = buf[o++] | (v << 8);
	h.type = v;
		
	s = header.index.size; o = header.index.offset; v = 0;
	while(s--) v = buf[o++] | (v << 8);
	h.index = v;

	s = header.length.size; o = header.length.offset; v = 0; 
	while(s--) v = buf[o++] | (v << 8);
	h.length = v;
	
	s = header.transaction.size; o = header.transaction.offset; v = 0; 
	while(s--) v = buf[o++] | (v << 8);
	h.transaction = v;

	if(h.version != ENTITY_MODEL_NUMBER) return false;
	if((h.type != FLASH_GLOBAL_FILE) && (h.type != FLASH_SOUND_FILE)) return false;
	if(h.length > ENTITY_FLASH_FILE_SIZE) return false;

	return true;
}

void EntityFlash::report(char *s) {

	TRACE("FLASH STATUS REPORT %s:\n", s);
	TRACE("\tNumber of file locations = %i\n", FLASH_N_FILES);
	TRACE("\tNumber of unused locations = %i\n", nSpare);
	TRACE("\tnextTransactionNumber = %i\n", nextTransactionNumber);
	TRACE("\tnVersions = %i\n", nVersions);
}














