/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityFlash.h								*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/

#ifndef __ENTITYFLASH_H__
#define __ENTITYFLASH_H__

//#define DEBUG_FLASH

#include <rtems.h>
#include <chameleon.h>
#include <flash.h>

#include "EntitySound.h"
#include "EntityGlobal.h"
#include "EntityPresets.h"
#include "EntityVersion.h"

#include "utils.h"

#define ENTITY_FLASH_MEMORY_SIZE 691200
#define ENTITY_FLASH_FILE_SIZE 2560
#define FLASH_MAX_SOUNDS SOUND_BANK_SIZE*5
#define FLASH_MAX_TRANSACTIONS 100000000
#define FLASH_N_FILES ENTITY_FLASH_MEMORY_SIZE/ENTITY_FLASH_FILE_SIZE
#define FLASH_GLOBAL_FILE 1
#define FLASH_SOUND_FILE 2

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// There are two numbers associated with each file:
//
// 	1. The location - which is the start address in flash memory of the file 
// 	when multiplied by the flash file size.
//
// 	2. The file index - which is usually derived from the sound number 
// 	ie A10 -> 9, B1 -> 40 etc, and is stored in the file header
//
// Transaction numbers:
//
// Each file has a 32 bit transaction number, enabling historical versions of each 
// file to be distinguished and recalled. There is no such thing as an erased file. 
// The most recent version of a file cannot be overwritten, except by a new version 
// of itself, old versions of files can be overwritten, according to a system of 
// priorities. This system works because the number of file locations is greater
// than the maximum number of named files.
//
// See constructor for file header structure.
//
// The minimum required sound file size is given by:
//
// 			    	(S - M) + (M + N) * 15
// 	File size (bytes) =     ----------------------  +  X  +  H  +  C
// 					  8
// where:
// 
// 	S = Size of matrix (= nparameters * modulation sources)
// 	M = Max number of non-zero matrix nodes
// 	N = nparameters
// 	X = extra sound data: 16 name bytes plus 256 tuning bytes
// 	H = header data: 10 bytes (check EntityFlash.cpp)
// 	C = CRC32 checksum: 4 bytes
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

class EntityFlash {

	struct HeaderItem {

		int	size;
		int	offset;
	};

	struct HeaderStructure {

		HeaderItem	idString;
		HeaderItem	version;
		HeaderItem	type;
		HeaderItem	index;
		HeaderItem	length;
		HeaderItem	transaction;

		int		size;
	};
	
	HeaderStructure	header;

	struct HeaderInfo {

		char		*idString;
		int		version;
		int		type;
		int		index;
		int		length;
		unsigned int	transaction;
	};

	int		nSpare;					// Number of spare file locations (eventually, always zero)
	bool		spare[FLASH_N_FILES];			// Which file locations are spare (unwritten or invalid)
	int		sound[FLASH_MAX_SOUNDS];		// Locations of most recent of each sound
	unsigned char	buf[ENTITY_FLASH_FILE_SIZE*2];		// General purpose buffer for manipulating header data
	unsigned char	buffer[ENTITY_FLASH_FILE_SIZE*2];	// General purpose buffer for files
	unsigned int	nextTransactionNumber;			// Transaction number for next file to be written
	int		globalLocation;				// Location of current (most recent) global file, or -1
	int		handle;					// Handle for Chameleon flash operations
	
	CRC32		crc;					// CRC32 checksum
	int 		checksumSize;

	int		newLocation();					// Identify location for new file (strategic)
	bool		getHdrInfo(int location, HeaderInfo &h);	// Get header data, checking validity
	bool		verify(int location);				// Fully verify file, using checksum
	int		getFileLength(int location);			// Get length specified in file header
	bool		writeFile(unsigned char*, HeaderInfo&, int);	// Write filedata in buffer according to header
	void		fixTransactionNumbers();			// The flash chip should fail before this gets called

	/* Sound file state (set from latest loaded sound file) */
	
	int		soundNumber;					// Number of last sound loaded
	int		location;					// Location of last sound loaded (as file number)
	int		version;					// Last version no. of sound loaded
	int		nVersions;					// Number of versions of current sound no.
	int		versionLocation[FLASH_MAX_SOUNDS];		// (Sortable) array of locations for sound versions
	unsigned int	versionTransaction[FLASH_MAX_SOUNDS];		// (Sortable) array of version transaction numbers 

	/* Preset sounds */

	EntityPresets*	presets;
	
public:

	EntityFlash();
	~EntityFlash();	

	int		validSoundNumber(int);				// Impose valid range for sound number	
	int		validSaveSoundNumber(int);			// Impose valid range for writable sound number	
	int		getnSounds();					// Max number of sounds, including presets 
	int		getnWritableSounds();				// Max number of (ie, writable) flash sounds 
	void		scan();						// Scan flash, updating state variables 
	void		setPriority(rtems_task_priority);		// Set priority level for background write task
	
	bool		load(EntityGlobal*);				// Load global file
	
	bool		save(EntityGlobal*);				// Save global file
	bool		save(rtems_unsigned8 *buffer, int bufSize);	// Save global from buffer (ie from midi)
	
	char		*snoopName(int n);				// Quickly get name for sound n

	bool		load(EntitySound*, int n);			// Read latest version of sound n
	int		load(int sn, rtems_unsigned8 *buffer);		// Read soundnumber sn into buffer (ie to send to midi)
	
	bool		save(EntitySound*);				   // Save sound (becomes latest version)
	bool		save(int n, rtems_unsigned8 *buffer, int bufSize); // Save sound from buffer (ie from midi)
	
	int		initVersionAccess(int n);			// Initialise version access for sound n
	bool		loadVersion(EntitySound*, int n);		// Read version n of sound
	bool		loadNextVersion(EntitySound*);			// Read next (chronological) version of sound
	bool		loadPreviousVersion(EntitySound*);		// Read previous (chronological) version of sound

	inline int	getversion() { return version; };		// Get current sound version (0 after initial load)
	inline int	getnVersions() { return nVersions; };		// Get no. of versions of last loaded sound

	void		report(char* = "");				// Print out status (for testing)
	inline int 	getnFiles() { return FLASH_N_FILES - nSpare; };	// Number of validated flash files on chip
};







#endif /* __ENTITYFLASH_H__ */
