/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityGlobal.cpp							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 ********************************************************************************/
  
#include "EntityGlobal.h"

//--------------------------------
//   Instantiate object counter
//--------------------------------

int GlobalPara::nGlobalParas = 0;

//----------------------------------------------------------------------
//
//   			Global Parameters
//
//----------------------------------------------------------------------

GlobalPara::GlobalPara(char* a, int min_, int max_, int defval_) { 

	char* n = name;
	while((*n++ = *a++));
	
	min = min_;
	max = max_;

	defval = defval_;
	useDefault = true;
	
	nGlobalParas++;
}

Status GlobalPara::incValue(int i) {

	int oldval = useDefault ? defval : value;
	int newval = validValue(oldval + i);

	if(newval != oldval) {
	
		useDefault = false;
		value = newval;
		return PARAMETER_CHANGED;
	}
	
	return PARAMETER_UNCHANGED;
}

int GlobalPara::validValue(int x) { 
	
	return x > max ? max : x < min ? min : x;
}

char* GlobalPara::getValStr(int &strLen) { 

	int v = useDefault ? defval : value;
	strLen = 8;	
	return i2s(v); 
}

GlobalCascadePara::GlobalCascadePara(char *n, int defval_):GlobalStringPara(n, N_CASCADE_MODES, defval_) {

	synq = getQueueId("SYNQ");

	static	char*	valueStrings[N_CASCADE_MODES];

	// Initialise translation LUT 

	int i = 0;
	
	for(int nm = 1; nm <= MAXIMUM_CASCADE; nm++) {		// number of machines
		for(int mi = 0; mi < nm; mi++, i++) {		// machine index
			
			lut[i].machineIndex = mi;
			lut[i].nMachines = nm;

			// ie: lut[0] = 0,1		"OFF"
			//     lut[1] = 0,2		"1/2"
			//     lut[2] = 1,2		"2/2"
			//     lut[3] = 0,3		"1/3"
			//     lut[4] = 1,3		"2/3"
			//     lut[5] = 2,3....		"3/3"....

			valueStrings[i] = new char[8];

			if(i == 0) {
				
				addString("OFF");
			}
			
			else {	
				
				// "1/2", "2/2", "1/3"... */
				sprintf(valueStrings[i], "%d/%d", mi+1, nm);	
				addString(valueStrings[i]);
			}
		}
	}

	if(i != N_CASCADE_MODES) TRACE("ERROR: Number of cascade modes doesn't match\n");
}

void GlobalCascadePara::updateSynth() {
	
	sendMessage(synq, RESET_VOICE_ASSIGN);	// Reset voice assigner on next note-on
}

Status GlobalCascadePara::incValue(int i) {

	int oldval = useDefault ? defval : value;
	int newval = validValue(oldval + i);

	if(newval != oldval) {
	
		useDefault = false;
		value = newval;

		updateSynth();
		
		return PARAMETER_CHANGED;
	}
	
	return PARAMETER_UNCHANGED;
}

void GlobalCascadePara::setValue(int x) { 
	
	useDefault = false; 
	value = validValue(x);
	updateSynth();
}

bool GlobalCascadePara::setDefault() { 
	
	useDefault = true; 
	updateSynth();
	return true; 
}

bool GlobalCascadePara::clearDefault() { 
	
	useDefault = false; 
	updateSynth();
	return true; 
}

void GlobalCascadePara::setValueToDefault() { 

	value = defval; 
	updateSynth();
	useDefault = false; 
}

GlobalTunePara::GlobalTunePara(char *n, int defval_):GlobalPara(n, -8192, 8191, defval_) { 

	synq = getQueueId("SYNQ");
}

GlobalCalibrateTuningPara::GlobalCalibrateTuningPara(char *n, int defval_):GlobalTunePara(n, defval_) { }

void GlobalTunePara::updateSynth() {

	mBuffer[0] = GLOBAL_TUNE;
	sendMessage(synq, mBuffer, 1);
}

Status GlobalTunePara::incValue(int i) {

	int oldval = useDefault ? defval : value;
	int newval = validValue(oldval + i);

	if(newval != oldval) {
	
		useDefault = false;
		value = newval;

		updateSynth();
		
		return PARAMETER_CHANGED;
	}
	
	return PARAMETER_UNCHANGED;
}

void GlobalTunePara::setValue(int x) { 
	
	useDefault = false; 
	value = validValue(x);
	updateSynth();
}

bool GlobalTunePara::setDefault() { 
	
	useDefault = true; 
	updateSynth();
	return true; 
}

bool GlobalTunePara::clearDefault() { 
	
	useDefault = false; 
	updateSynth();
	return true; 
}

void GlobalTunePara::setValueToDefault() { 

	value = defval; 
	updateSynth();
	useDefault = false; 
}

int GlobalTunePara::translate() { 

	int v = useDefault ? defval : value;
	
	// NB value stored is offset in hundredths of Hz from A = 440Hz
	// - the translated value needs to be a pitch offset and 
	// this will be done in EntitySynth, after the master tuning and 
	// calibration tuning frequency offsets have been added together.
	// This strategy is due to the amount of calculation required 
	// (a base 2 logarithm).

	return v; 
}

char *GlobalTunePara::getValStr(int &strLen) { 

	static char s[16];
	
	int v = useDefault ? defval : value;
	int thz = v + 44000;
	
	sprintf(s, "A = %-.3i.%-.2i Hz", thz / 100, thz % 100);

	strLen = 13;
	
	return s;
}

char *GlobalCalibrateTuningPara::getValStr(int &strLen) { 

	static char s[16];
	
	int v = useDefault ? defval : value;
	
	if(v < 0) { v *= -1; sprintf(s, "-%2.2i.%2.2i Hz", v / 100, v % 100); }
	else sprintf(s, " %2.2i.%2.2i Hz", v / 100, v % 100);

	strLen = 10;
	
	return s;
}

GlobalSavePara::GlobalSavePara(char *n, SaveType saveType_):GlobalPara(n, 0, 1) {

	saveType = saveType_;
}
	
char *GlobalSavePara::getValStr(int &strLen) {

	strLen = 0; 
	return 0;
}
	
GlobalStringPara::GlobalStringPara(char *n, int a, int defval_):GlobalPara(n, 0, a-1, defval_) {

	paraString = new char* [a+1];
	paraLength = new int [a+1];

	addIndex = 0;
}

GlobalStringPara::~GlobalStringPara() {

	for(int x = 0; x <= max; x++) {
		delete paraString[x];
	}

	delete [] paraString;
	delete [] paraLength;
}

void GlobalStringPara::addString(char *s) {

	int i = 0; 
	char *q = s;
	while(*q++) i++;
	paraLength[addIndex] = i;
	paraString[addIndex] = s;

	addIndex++;
}

char *GlobalStringPara::getValStr(int &strLen) {

	int v = useDefault ? defval : value;
	strLen = paraLength[v];
	return paraString[v];
}

GlobalReceivePara::GlobalReceivePara(char *n, int defval_):GlobalStringPara(n, 2, defval_) {

	addString("Local");
	addString("Midi + local");
}

GlobalTChanPara::GlobalTChanPara(char *n, int defval_):GlobalStringPara(n, 16, defval_) {

	addString(" 1");
	addString(" 2");
	addString(" 3");
	addString(" 4");
	addString(" 5");
	addString(" 6");
	addString(" 7");
	addString(" 8");
	addString(" 9");
	addString("10");
	addString("11");
	addString("12");
	addString("13");
	addString("14");
	addString("15");
	addString("16");
}

GlobalDevIdPara::GlobalDevIdPara(char *n, int defval_):GlobalStringPara(n, 17, defval_) {

	addString(" 1");
	addString(" 2");
	addString(" 3");
	addString(" 4");
	addString(" 5");
	addString(" 6");
	addString(" 7");
	addString(" 8");
	addString(" 9");
	addString("10");
	addString("11");
	addString("12");
	addString("13");
	addString("14");
	addString("15");
	addString("16");
	addString("ANY");
}

int GlobalDevIdPara::translate() { 

	int v = useDefault ? defval : value;
	
	if(v == 16) return 127;
	
	return v; 
}

GlobalRChanPara::GlobalRChanPara(char *n, int defval_):GlobalStringPara(n, 17, defval_) {

	addString(" 1");
	addString(" 2");
	addString(" 3");
	addString(" 4");
	addString(" 5");
	addString(" 6");
	addString(" 7");
	addString(" 8");
	addString(" 9");
	addString("10");
	addString("11");
	addString("12");
	addString("13");
	addString("14");
	addString("15");
	addString("16");
	addString("ALL");	// Omni
}

GlobalTRPara::GlobalTRPara(char *n, int defval_):GlobalStringPara(n, 4, defval_) {

	addString("Off");
	addString("Transmit");
	addString("Receive");
	addString("Transmit+receive");
}

bool GlobalTRPara::transmit() { return getValue() & 1; } 

bool GlobalTRPara::receive() { return getValue() > 1; } 

GlobalOnOffPara::GlobalOnOffPara(char *n, int defval_):GlobalStringPara(n, 2, defval_) {

	addString("Off");
	addString("On");
}

GlobalStereoPara::GlobalStereoPara(char *n, int defval_):GlobalStringPara(n, 2, defval_) {

	addString("Mono");
	addString("Stereo");
}

void GlobalMidiCtrPara::addString(char *s, int v) {

	int i = 0; 
	char *q = s;
	while(*q++) i++;
	paraLength[addIndex] = i;
	paraString[addIndex] = s;
	paraValue[addIndex] = v;

	addIndex++;
}

GlobalMidiCtrPara::GlobalMidiCtrPara(char *n, int defval_, char ctrBit_, char *ctrLUT_):GlobalStringPara(n, 128, defval_) {

	paraValue = new int [128];
	for(int x = 0; x < 128; x++) paraValue[x] = 0;

	// global ctrLUT 
	
	ctrBit = ctrBit_;
	ctrLUT = ctrLUT_;
	
	ctrLUT[paraValue[defval]] |= ctrBit;	
}

void GlobalMidiCtrPara::bTrace(int v) {

	TRACE("ctrLUT[%i] = \t", paraValue[v]);

	for(int xx = 1; xx < 256; xx <<= 1) {

		if(xx & ctrLUT[paraValue[value]]) TRACE("1");
		else TRACE("0");
	}

	TRACE("\n");
}

void GlobalMidiCtrPara::setValue(int x) { 
	
	ctrLUT[paraValue[value]] &= ~ctrBit;

	useDefault = false; 
	value = validValue(x);

	ctrLUT[paraValue[value]] |= ctrBit;
}

bool GlobalMidiCtrPara::setDefault() { 
	
	useDefault = true; 

	ctrLUT[paraValue[value]] &= ~ctrBit;
	ctrLUT[paraValue[defval]] |= ctrBit;
	
	return true;
}

bool GlobalMidiCtrPara::clearDefault() { 
	
	useDefault = false; 

	ctrLUT[paraValue[defval]] &= ~ctrBit;
	ctrLUT[paraValue[value]] |= ctrBit;

	return true;
}

void GlobalMidiCtrPara::setValueToDefault() { 
	
	ctrLUT[paraValue[value]] &= ~ctrBit;
	ctrLUT[paraValue[defval]] |= ctrBit;	

	value = defval; 
	useDefault = false;
};

Status GlobalMidiCtrPara::incValue(int i) {

	int oldval = useDefault ? defval : value;
	int newval = validValue(oldval + i);

	if(newval != oldval) {
	
		ctrLUT[paraValue[oldval]] &= ~ctrBit;
		ctrLUT[paraValue[newval]] |= ctrBit;
		
		useDefault = false;
		value = newval;

		return PARAMETER_CHANGED;
	}
	
	return PARAMETER_UNCHANGED;
}

GlobalMidiCtrPara::~GlobalMidiCtrPara() {

	for(int x = 0; x <= max; x++) {
		delete paraString[x];
	}

	delete [] paraString;
	delete [] paraLength;
	delete [] paraValue;
}

GlobalMidiCtr7Para::GlobalMidiCtr7Para(char *n, int defval_, char ctrBit_, char *ctrLUT_):GlobalMidiCtrPara(n, defval_, ctrBit_, ctrLUT_) {

	// Midi Controllers
	
	addString("Aftertouch", 32);	// Dummy value to avoid problems with ctrLUT 
					// - ctr #32 is LSB of bank change; unused
	addString("1 Modulation", 1);
	addString("2 Breath Ctrl", 2);
	addString("3", 3);
	addString("4 Foot Pedal", 4);
	addString("5 Portamento", 5);
	addString("7 Volume", 7);
	addString("8 Balance", 8);
	addString("9", 9);
	addString("10 Pan", 10);
	addString("11 Expression", 11);
	addString("12 Effect Ctrl 1", 12);
	addString("13 Effect Ctrl 2", 13);
	addString("14", 14);
	addString("15", 15);
	addString("16 Slider 1", 16);
	addString("17 Slider 2", 17);
	addString("18 Slider 3", 18);
	addString("19 Slider 4", 19);
	addString("20", 20);
	addString("21", 21);
	addString("22", 22);
	addString("23", 23);
	addString("24", 24);
	addString("25", 25);
	addString("26", 26);
	addString("27", 27);
	addString("28", 28);
	addString("29", 29);
	addString("30", 30);
	addString("31", 31);	

	addString("64 Sustain Pedal", 64);
	addString("65 Portamento Sw", 65);
	addString("66 Sostenuto Ped", 66);
	addString("67 Soft Pedal", 67);
	addString("68 Legato Pedal", 68);
	addString("69 Hold 2 Pedal", 69);
	addString("70 Variation", 70);
	addString("71 Timbre", 71);
	addString("72 Release", 72);
	addString("73 Attack", 73);
	addString("74 Brightness", 74);
	addString("75 Control 6", 75);
	addString("76 Control 7", 76);
	addString("77 Control 8", 77);
	addString("78 Control 9", 78);
	addString("79 Control 10", 79);
	addString("80 GP Button 1", 80);
	addString("81 GP Button 2", 81);
	addString("82 GP Button 3", 82);
	addString("83 GP Button 4", 83);
	addString("84", 84);
	addString("85", 85);
	addString("86", 86);
	addString("87", 87);
	addString("88", 88);
	addString("89", 89);
	addString("90", 90);
	addString("91 Effects Level", 91);
	addString("92 Tremolo Level", 92);
	addString("93 Chorus Level", 93);
	addString("94 Celeste Level", 94);
	addString("95 Phaser Level", 95);

	max = addIndex - 1; // saves counting manually
}

GlobalMidiCtr14Para::GlobalMidiCtr14Para(char *n, int defval_, char ctrBit_, char *ctrLUT_) : GlobalMidiCtrPara(n, defval_, ctrBit_, ctrLUT_) {

	// Midi Controllers
	
	addString("1 Modulation", 1);
	addString("2 Breath Ctrl", 2);
	addString("3", 3);
	addString("4 Foot Pedal", 4);
	addString("5 Portamento", 5);
	addString("7 Volume", 7);
	addString("8 Balance", 8);
	addString("9", 9);
	addString("10 Pan", 10);
	addString("11 Expression", 11);
	addString("12 Effect Ctrl 1", 12);
	addString("13 Effect Ctrl 2", 13);
	addString("14", 14);
	addString("15", 15);
	addString("16 Slider 1", 16);
	addString("17 Slider 2", 17);
	addString("18 Slider 3", 18);
	addString("19 Slider 4", 19);
	addString("20", 20);
	addString("21", 21);
	addString("22", 22);
	addString("23", 23);
	addString("24", 24);
	addString("25", 25);
	addString("26", 26);
	addString("27", 27);
	addString("28", 28);
	addString("29", 29);
	addString("30", 30);
	addString("31", 31);	
	
	max = addIndex - 1; // saves counting manually
}

//-----------------------------------------------------------------------------------------------
//
// 					Global
//
//-----------------------------------------------------------------------------------------------

EntityGlobal::EntityGlobal() { 

	for(int x = 0; x < 128; x++) ctrLUT[x] = 0; 	// Controller mapping LUT
	
	GlobalPara::countZero();
	GlobalPara **p = globalPara;
	
	// - - - - - - - - - - - - - - - - - - -
	// Hidden parameters 
	
	/* Hidden Global Parameters, not accessible to user */
	*p++ = entitymn = new GlobalPara("Entity Model", -8192, +8191, -1);
	*p++ = majorvrn = new GlobalPara("Major version", -8192, +8191, -1);
	*p++ = minorvrn = new GlobalPara("Minor version", -8192, +8191, -1);
	*p++ = testvrsn = new GlobalPara("Test version", -8192, +8191, -1);
	
	/* Hidden Global Parameter, user-accessible through obscure keys */
	*p++ = calitune = new GlobalCalibrateTuningPara("CALIBRATE TUNING", 0);

	/* Hidden Global Parameter, user-changeable through EDIT+SHIFT */
	*p++ = fastedit = new GlobalPara("Fast edit", 0, 1, 1);

	/* Hidden Global Parameter, user-changeable through SHIFT while in EDIT mode */
	*p++ = shiftkey = new GlobalPara("Shift key", 0, 1, 0);

	// - - - - - - - - - - - - - - - - - - -
	// Directly-editable global parameters 
	
	baseParaIndex = GlobalPara::getCount();
	
	*p++ = midirchn = new GlobalRChanPara    ("MIDI RX CHANNEL ", 0);
	*p++ = miditchn = new GlobalTChanPara    ("MIDI TX CHANNEL ", 0);
	*p++ = mididvid = new GlobalDevIdPara	 ("MIDI DEVICE ID  ", 16);
	*p++ = midiecho = new GlobalOnOffPara    ("   MIDI THRU    ", 0);
	*p++ = midiptch = new GlobalPara         ("MIDI PITCH BEND ", -12, 12, 2);
	*p++ = midiedit = new GlobalTRPara       (" MIDI NRPN DATA ", 3);
	*p++ = midivolr = new GlobalTRPara       ("  MIDI VOLUME   ", 2);
	*p++ = midiprgr = new GlobalTRPara       ("MIDI PROG CHANGE", 2);
	
	*p++ = midictrl = new GlobalTRPara       ("MIDI CONTROLLERS", 3);
	
	*p++ = midictr1 = new GlobalMidiCtr7Para (" MIDI CTR1 RTC  ", 15, 0x01, ctrLUT);
	*p++ = midictr2 = new GlobalMidiCtr7Para (" MIDI CTR2 RTC  ", 16, 0x02, ctrLUT);
	*p++ = midictr3 = new GlobalMidiCtr7Para (" MIDI CTR3 RTC  ", 17, 0x04, ctrLUT);
	*p++ = midimodl = new GlobalMidiCtr7Para (" MIDI MMOD RTC  ", 1,  0x08, ctrLUT);
	
	*p++ = encrctr1 = new GlobalOnOffPara 	 (" ENCODER \x01 CTR1 ", 0);
	*p++ = midicasc = new GlobalCascadePara	 ("  MIDI CASCADE  ", 0);
	*p++ = glbltune = new GlobalTunePara	 (" MASTER TUNING  ", 0);
	
	*p++ = new GlobalSavePara		 ("MIDI SAVE CONFIG", MIDI_SAVE_CONFIG);
	*p++ = new GlobalSavePara		 (" MIDI SAVE ALL  ", MIDI_SAVE_ALL);
	*p++ = new GlobalSavePara		 (" MIDI SAVE BANK ", MIDI_SAVE_BANK);
	*p++ = new GlobalSavePara		 ("MIDI SAVE SOUND ", MIDI_SAVE_SOUND);

	nVisibleParas = GlobalPara::getCount() - baseParaIndex;

	/* Init pot LUT for fast edits */
	initPotLUT(nVisibleParas, potLUT);
	
	nParas = GlobalPara::getCount();

	currentParameterIndex = baseParaIndex;
	
	for(int x = 0; x < nParas; x++) globalPara[x]->translate();
}

bool EntityGlobal::parameterInc(int x) {

	int newParameterIndex = currentParameterIndex + x;
	
	if(newParameterIndex < baseParaIndex) newParameterIndex = baseParaIndex;
	if(newParameterIndex >= nParas) newParameterIndex = nParas - 1;
       
	if(currentParameterIndex == newParameterIndex) return false;
	currentParameterIndex = newParameterIndex;

	return true;
}

int EntityGlobal::getCurrentParameterIndex() {

	return currentParameterIndex;
}

bool EntityGlobal::setParameterByPot(int p) {

	int newParameterIndex = potLUT[p & 0xff] + baseParaIndex;
	if(newParameterIndex == currentParameterIndex) return false;

	currentParameterIndex = newParameterIndex;
	return true;
}

int EntityGlobal::setParameterByIndex(int x) {

	if(x < baseParaIndex) return currentParameterIndex = 0;
	if(x >= nParas) return currentParameterIndex = nParas - 1;
       
	currentParameterIndex = x;

	return true;
}

bool EntityGlobal::selectFirstParameter() {

	if(currentParameterIndex == baseParaIndex) return false;
	else currentParameterIndex = baseParaIndex;
	return true;
}

bool EntityGlobal::selectLastParameter() {

	if(currentParameterIndex == nParas - 1) return false;
	else currentParameterIndex = nParas - 1;
	return true;
}

GlobalPara* EntityGlobal::getCurrentParameter() {

	return globalPara[currentParameterIndex];
}

GlobalPara* EntityGlobal::getParameterByIndex(int index) {

	return globalPara[index];
}

GlobalPara* EntityGlobal::getCalibrateTuningPara() {

	return calitune;
}

int EntityGlobal::pack(unsigned char *buf) {

	// Store each value (whether it is a value-para or not)
	// in a pair of bytes, MSB first. Plenty of room.

	int x; for(x = 0; x < nParas; x++) {

		int v = globalPara[x]->getValue();
		
		*buf++ = (unsigned char) (v >> 8);
		*buf++ = (unsigned char) v;
	}
		
	return x * 2;	// return number of bytes	
}

bool EntityGlobal::unpack(unsigned char *buf) {

	for(int x = 0; x < nParas; x++) {

		int vH = (signed char) *buf++;
		int vL = *buf++;

		globalPara[x]->setValue((vH << 8) | vL);
	}

	return true;
}

void EntityGlobal::loadDefaultValues() {

	for(int x = 0; x < nParas; x++) {

		globalPara[x]->setValueToDefault();
	}
}

bool EntityGlobal::midiDeviceReceive(int devId) { 

	int synthDevId = mididvid->translate();
	if(synthDevId == 127) return true;
	if(devId == synthDevId) return true;
	return false;
}

bool EntityGlobal::midiReceive(int chan) {

	if(midirchn->getValue() == 16) return true; 	// OMNI
	if(midirchn->getValue() == chan) return true; 	// Channels match
	return false;					// Channels don't match
}

bool EntityGlobal::matchesVersion(int entityModelNumber, int majorVersionNumber, int minorVersionNumber, int testVersionNumber) {

	if(entityModelNumber != entitymn->getValue()) return false;
	if(majorVersionNumber != majorvrn->getValue()) return false;
	if(minorVersionNumber != minorvrn->getValue()) return false;
	if(testVersionNumber != testvrsn->getValue()) return false;

	return true;
}

void EntityGlobal::setVersion(int entityModelNumber, int majorVersionNumber, int minorVersionNumber, int testVersionNumber) {

	entitymn->setValue(entityModelNumber);
	majorvrn->setValue(majorVersionNumber);
	minorvrn->setValue(minorVersionNumber);
	testvrsn->setValue(testVersionNumber);
}

bool EntityGlobal::fastEdit() { return (bool) fastedit->getValue(); } 
		
bool EntityGlobal::toggleFastEdit() {

	if(fastedit->getValue()) { fastedit->setValue(0); return false; }
	fastedit->setValue(1); 
	return true; 
}

bool EntityGlobal::shiftKey() { return (bool)shiftkey->getValue(); } 
		
bool EntityGlobal::toggleShiftKey() {

	if(shiftkey->getValue()) { shiftkey->setValue(0); return false; }
	shiftkey->setValue(1); 
	return true; 
}
