/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityMessageHandler.cpp						*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/
  
#include "EntityMessageHandler.h"

bool EntityMessageHandler::createQueue(char* name, int nMaxMessages, int maxMessageSize) {

	if(maxMessageSize > MAX_MESSAGE_SIZE) {	TRACE("ERROR: Message size too large\n"); return false; }

	rtems_id q_id;	// Just discard the queue id - tasks that need access create their own queue ids
	
	status = rtems_message_queue_create(	rtemsName(name), 
						nMaxMessages, 
						maxMessageSize, 
						RTEMS_FIFO | RTEMS_LOCAL, 
						&q_id
	);
	
	if (status != RTEMS_SUCCESSFUL)	{ TRACE("ERROR: cannot create new queue.\n"); return FALSE; }

	return true;
}

bool EntityMessageHandler::deleteQueue(char* name) {

	rtems_id q_id = getQueueId(name);
	
	status = rtems_message_queue_delete(q_id);
	
	if (status != RTEMS_SUCCESSFUL)	{ TRACE("ERROR: cannot delete queue.\n"); return FALSE; }

	return true;
}

bool EntityMessageHandler::clearQueue(rtems_id q_id) {

	rtems_unsigned32 n;
	
	status = rtems_message_queue_flush(q_id, &n);
	
	if (status != RTEMS_SUCCESSFUL)	{ TRACE("ERROR: cannot flush queue.\n"); return false; }

	return true;
}

rtems_name EntityMessageHandler::rtemsName(char* name) {

	return rtems_build_name(name[0], name[1], name[2], name[3]);
}

rtems_id EntityMessageHandler::getQueueId(char* name) {

	rtems_id q_id = 0;

	status = rtems_message_queue_ident(rtemsName(name), RTEMS_SEARCH_ALL_NODES, &q_id);	
	if(status != RTEMS_SUCCESSFUL) TRACE("ERROR: Cannot get queue id in EntityMessageHandler::getQueueId()\n");

	return q_id;
}

int EntityMessageHandler::getMessage(rtems_id q_id) {

	rtems_unsigned32 nBytes;
       
	status = rtems_message_queue_receive(q_id, mBuffer, &nBytes, RTEMS_WAIT, RTEMS_NO_TIMEOUT);
	if (status != RTEMS_SUCCESSFUL) return 0;
	
	return nBytes;
}

int EntityMessageHandler::getMessage(rtems_id q_id, void *m) {

	rtems_unsigned32 nBytes;
       
	status = rtems_message_queue_receive(q_id, m, &nBytes, RTEMS_WAIT, RTEMS_NO_TIMEOUT);
	if (status != RTEMS_SUCCESSFUL) return 0;
	
	return nBytes;
}

int EntityMessageHandler::sendMessage(rtems_id q_id, char value) {

	return sendMessage(q_id, &value, 1);
}

int EntityMessageHandler::sendUrgentMessage(rtems_id q_id, char value) {

	return sendMessage(q_id, &value, 1);
}

int EntityMessageHandler::sendMessage(rtems_id q_id, void *tBuffer, int nBytes) { 

	// Return number of bytes successfully transmitted
	
	status = rtems_message_queue_send(q_id, tBuffer, nBytes);

	switch(status) {

		case RTEMS_SUCCESSFUL :		return nBytes;
		case RTEMS_TOO_MANY :		TRACE("LIMIT: Too many messages.\n"); break;

		case RTEMS_INVALID_ID :		TRACE("ERROR: Invalid queue id for sendMessage()\n"); break;
		case RTEMS_INVALID_SIZE :	TRACE("ERROR: Invalid message size for sendMessage()\n"); break;
		case RTEMS_UNSATISFIED :	TRACE("ERROR: Out of message buffers in sendMessage()\n"); break;
		
		default :			TRACE("ERROR: Couldn't write to message queue\n"); break;
	}

	return 0;
}

int EntityMessageHandler::sendUrgentMessage(rtems_id q_id, void *tBuffer, int nBytes) { 

	// Return number of bytes successfully transmitted
	
	status = rtems_message_queue_urgent(q_id, tBuffer, nBytes);

	switch(status) {

		case RTEMS_SUCCESSFUL :		return nBytes;
		case RTEMS_TOO_MANY :		TRACE("LIMIT: Too many messages.\n"); break;

		case RTEMS_INVALID_ID :		TRACE("ERROR: Invalid queue id for sendMessage()\n"); break;
		case RTEMS_INVALID_SIZE :	TRACE("ERROR: Invalid message size for sendMessage()\n"); break;
		case RTEMS_UNSATISFIED :	TRACE("ERROR: Out of message buffers in sendMessage()\n"); break;
		
		default :			TRACE("ERROR: Couldn't write to message queue\n"); break;
	}

	return 0;
}

EntityMessageHandler::EntityMessageHandler() { }
EntityMessageHandler::~EntityMessageHandler() { }

