/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntitySoundMode.cpp							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/

#include "EntitySoundMode.h"

EntitySoundMode::EntitySoundMode(EntityPanel *panel_, EntitySoundManager *soundManager_, EntityGlobal *global_) : EntityMode(panel_) {

	soundManager = soundManager_;
	global = global_;
}

Mode EntitySoundMode::enter() {

	soundManager->update();

	panel->clearLEDs(PANEL01_LED_EDIT, PANEL01_LED_CTRL1, PANEL01_LED_CTRL2, PANEL01_LED_CTRL3);

	panel->setEncoderAcceleration(0);

	// Possibly set encoder as CTR1 fine control 
	if(global->encrCtr1()) panel->setEncoderForModulation(); 
	
	if(!soundManager->getSound()) TRACE("ERROR - no sound in EntitySoundMode::enter()\n");

	number = soundManager->getNumber();

	processEvents();
	while(soundManager->loading()) rtems_task_wake_after(50);

	panel->setEncoderForData(); // Always use encoder for data entry in other modes
	
	return rMode;
}

void EntitySoundMode::edit() { rMode = EDIT; }				

void EntitySoundMode::partUp() { 
	
	panel->repeat(AUTO_VERSION_UP, 825, 50);
	if(soundManager->versionUp()) displayVersionUp();
}

void EntitySoundMode::partDown() {
	
	panel->repeat(AUTO_VERSION_DOWN, 825, 50);
	if(soundManager->versionDown()) displayUpdate();
}

void EntitySoundMode::pageUp() { rMode = OUTPUT; }
void EntitySoundMode::pageDown() { rMode = INPUT; }

void EntitySoundMode::paramUp() { incrementEvent(soundManager->getBankSize()); }
void EntitySoundMode::paramDown() { decrementEvent(soundManager->getBankSize()); }

void EntitySoundMode::valueUp() {

	panel->repeat(AUTO_VALUE_UP, 825, 50);
	incrementEvent();
}

void EntitySoundMode::valueDown() {

	panel->repeat(AUTO_VALUE_DOWN, 825, 50);
	decrementEvent();
}

void EntitySoundMode::autoEvent(int event) {

	switch(event) {

		case NEW_SOUND_NAME_READY :	displayNewName(); break;
		case NEW_SOUND_READY :		displayBullet(); break;
		case AUTO_VALUE_UP :		incrementEvent(); break;
		case AUTO_VALUE_DOWN :		decrementEvent(); break;
		case AUTO_VERSION_UP :		if(soundManager->versionUp()) displayVersionUp(); break;
		case AUTO_VERSION_DOWN :	if(soundManager->versionDown()) displayUpdate(); break;
		case CLEAR_VERSION_INFO :	panel->clearLine(1); break;
		case MIDI_PROG_CHANGE :		midiProgChange(); break;
	}
}

void EntitySoundMode::midiProgChange() {

	int soundBank = mBuffer[1];
	int userSoundNumber = mBuffer[2];
 
	int bankSize = soundManager->getBankSize();
	
	/* If midi requests sound >= bank size, ignore request */ 
	if(userSoundNumber >= bankSize) return;
	
	/* If soundBank = -1, userSoundNumber is offset within current bank */
	if(soundBank < 0) soundBank = number / soundManager->getBankSize();

	number = soundBank * soundManager->getBankSize() + userSoundNumber;

	if(number < 0) number = 0;
	if(number > soundManager->getMaxNumber()) number = soundManager->getMaxNumber();

	soundManager->bgLoadSound(number);

	displayValueUpdate();
}

Status EntitySoundMode::incrementValue(int i) {

	int oldNumber = number;
	number += i;
				
	if(number < 0) number = 0;
	else if(number > soundManager->getMaxNumber()) number = soundManager->getMaxNumber();

	if(oldNumber == number) return PARAMETER_UNCHANGED;

	// The following message requests the synth object to
	// send a midi prog change command if appropriate
	
	sm[0] = PROGRAM_CHANGE;
	sm[1] = number / soundManager->getBankSize();
	sm[2] = number % soundManager->getBankSize();

	sendMessage(synq, sm, 3);
	
	// Initiate background loading sequence
	
	soundManager->bgLoadSound(number);

	return PARAMETER_CHANGED;
}

void EntitySoundMode::displayUpdate() {

	char 	line[PANEL_LCD_MAX_LINE_LEN + 7];

	char*		s;	// string to copy
	int 		p;	// position
		
	s = soundManager->getNumberStr(number);
	p = 0; while(*s) line[p++] = *s++;

	line[p++] = soundManager->getBullet();

	s = soundManager->getNameStr();
	while((line[p++] = *s++));

	panel->clear();
	panel->print(0, 0, line);

	/* Version no. prop. to age, starting zero, or -1 if edited.
	 * For vn > 0, display with high version n. as more recent */

	int v = soundManager->getversion(); // -1..(nVersions-1)

	if(v > 0) { // if not edited or most recent version of sound
		
		panel->clearDelay();
		int nv = soundManager->getnVersions();
		sprintf(line, "version %i of %i", nv - v, nv);
		panel->print(1, 0, line);
	}
}

void EntitySoundMode::displayValueUpdate() {

	/* First stage of multi-stage loading */
	/* (display sound bank and number) */
	
	panel->clear();
	char *s = soundManager->getNumberStr(number);
	numberLength = strlen(s);
	
	panel->print(0, 0, s);
}

void EntitySoundMode::displayNewName() {

	/* Intermediate stage of multi-stage loading */
	
	panel->print(0, numberLength + 1, soundManager->getNewNameStr());
}

void EntitySoundMode::displayBullet() {

	/* Final stage of multi-stage loading */

	static char s[2] = {0, 0};
	*s = soundManager->getBullet();	
	panel->print(0, numberLength, s);
}

void EntitySoundMode::displayVersionUp() {

	char 	line[PANEL_LCD_MAX_LINE_LEN + 7];

	/* Ordinary display, inc versions > 0 */

	displayUpdate();	
	
	/* Optional timed display for version 0 */

	if(soundManager->getversion() == 0) {

		panel->delay(CLEAR_VERSION_INFO, 2500, 0);
		int nv = soundManager->getnVersions();
		sprintf(line, "version %i of %i", nv, nv);
		panel->print(1, 0, line);
	}
}	

