/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntitySynth.cpp							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005                                      *
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/
  
#include "EntitySynth.h"
#include "dsp/dsp_code.h"

//#define DEBUG_SYNTH

// EntitySynth is the "junction box" that connects midi, dsp and UI. It receives messages
// on SYNQ (and, during sound selection, SSTQ). It also makes use of the voice assigner 
// defined in EntityVoiceAssigner.cpp/h to translate midi note events into dsp events.

EntitySynth::EntitySynth(EntityGlobal* global_, ParaStructure* paras_, EntityMidi* midi_, EntityDsp* dsp_, EntityVoiceAssigner* assign_) {

	global = global_;
	assign = assign_;
	paras = paras_;
	midi = midi_;
	dsp = dsp_;

	synq = getQueueId("SYNQ");
	ctrq = getQueueId("CTRQ");
	sstq = getQueueId("SSTQ");

	paraForEdit = 0;
	sourceForEdit = -1;
	dspParaIndex = -1;
	
	pendingSound = 0;	
	activeSound = 0;

	expression = 16383;		// Midi expression (applied with volume)
	volume = 16383;			// Midi volume
	note = 0;			// Last note (0..127)
	
	em = 0;				// Encoder modulation (when used for CTR1) 

	// Initialise LUT of vectors for responding to incoming synth messages
	
	for(int x = 0; x < N_SYNTH_EVENT_TYPES; x++) event[x] = &EntitySynth::NullEvent;

	event[PANEL_POTENTIOMETER]		= &EntitySynth::PanelPotentiometer;
	event[PANEL_ENCODER_MOD]		= &EntitySynth::PanelEncoderMod;
	event[PARAMETER_SELECT]			= &EntitySynth::ParameterSelect;
	event[PARAMETER_MODIFY]			= &EntitySynth::ParameterModify;
	event[MIDI_NOTE_ON]			= &EntitySynth::MidiNoteOn;
	event[MIDI_NOTE_OFF]			= &EntitySynth::MidiNoteOff;
	event[MIDI_SUSTAIN]			= &EntitySynth::MidiSustain;
	event[MIDI_CONTROLLER]			= &EntitySynth::MidiController;
	event[MIDI_PITCH]			= &EntitySynth::MidiPitch;
	event[MIDI_VOLUME]			= &EntitySynth::MidiVolume;
	event[MIDI_PAN]				= &EntitySynth::MidiPan;
	event[MIDI_EXPRESSION]			= &EntitySynth::MidiExpression;
	event[MIDI_ALL_SOUND_OFF]		= &EntitySynth::MidiAllSoundOff;
	event[MIDI_ALL_NOTES_OFF]		= &EntitySynth::MidiAllNotesOff;
	event[MIDI_RESET]			= &EntitySynth::MidiReset;
	event[GLOBAL_TUNE]			= &EntitySynth::GlobalTune;
	event[NOTE_EVENT]			= &EntitySynth::NoteEvent;
	event[ALL_NOTES_OFF]			= &EntitySynth::AllNotesOff;
	event[NEW_SOUND_READY]			= &EntitySynth::NewSoundReady;
	event[READ_INPUT_LEVELS]		= &EntitySynth::ReadInputLevels;
	event[READ_OUTPUT_LEVELS]		= &EntitySynth::ReadOutputLevels;
	event[CLEAR_DSP_LEVELS]			= &EntitySynth::ClearDSPLevels;
	event[READ_DSP_MIPS]			= &EntitySynth::ReadDSPMIPS;
	event[READ_DSP_DEBUG_DATA]		= &EntitySynth::ReadDSPDebugData;
	event[RESET_DSP_DEBUG_ADDRESS]		= &EntitySynth::ResetDSPDebugAddress;
	event[MIDI_NRPN_HI_RES_UPDATE]		= &EntitySynth::MidiNRPNHiResUpdate;
	event[MIDI_NRPN_LO_RES_UPDATE]		= &EntitySynth::MidiNRPNLoResUpdate;
	event[MIDI_NRPN_INCREMENT]		= &EntitySynth::MidiNRPNIncrement;
	event[MIDI_NRPN_DECREMENT]		= &EntitySynth::MidiNRPNDecrement;
	event[PROGRAM_CHANGE]			= &EntitySynth::ProgramChange;
	event[MIDI_CLOCK]			= &EntitySynth::MidiClock;
	event[MIDI_LFO_SYNC_SUSTAIN_ON]		= &EntitySynth::MidiLFOSyncSustainOn;
	event[MIDI_LFO_SYNC_SUSTAIN_OFF]	= &EntitySynth::MidiLFOSyncSustainOff;
	event[MIDI_LFO_SYNC_PHASE]		= &EntitySynth::MidiLFOSyncPhase;
	event[RESET_VOICE_ASSIGN]		= &EntitySynth::ResetVoiceAssign;
	event[START_STACK_EDIT]			= &EntitySynth::StartStackEdit;
	event[END_STACK_EDIT]			= &EntitySynth::EndStackEdit;
}

void EntitySynth::Run() { while(1) { getMessage(synq); (this->*event[mBuffer[0]])(); } }

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// 			      SYNTH DECODE VECTORS
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

void EntitySynth::ResetVoiceAssign() 	  { assign->resetVoiceAssign(); }
void EntitySynth::StartStackEdit() 	  { assign->startStackEdit(); }
void EntitySynth::EndStackEdit() 	  { assign->endStackEdit(); }

void EntitySynth::MidiClock() 	 	  { dsp->command(DSPP_HV_MIDI_CLOCK); }
void EntitySynth::MidiLFOSyncSustainOn()  { dsp->command(DSPP_HV_MIDI_SYNC_SUSTAIN_ON); }
void EntitySynth::MidiLFOSyncSustainOff() { dsp->command(DSPP_HV_MIDI_SYNC_SUSTAIN_OFF); }
void EntitySynth::MidiLFOSyncPhase() 	  { dsp->update( DSPP_HV_MIDI_SYNC_PHASE, mBuffer[1]); }

void EntitySynth::PanelPotentiometer() {

	EntitySound *s = pendingSound ? pendingSound : activeSound;

	int v = mBuffer[2];

	switch(mBuffer[1]) {		

		// Pots are always positive 24-bit signed fractions

		case PANEL01_POT_VOLUME : 
						
			dsp->update(DSPP_HV_MASTER_VOL, v << 16); 
			if(global->midiVolTransmit()) midi->newCtrlChangeEvent(7, v);

			break;
			
		case PANEL01_POT_CTRL1 	: 
								  
			s->setPot(0, v); 
			em = v << 7;
			dsp->update(DSPP_HV_CTR1, v << 16); 
			if(global->midiCtrlTransmit()) midi->newCtrlChangeEvent(global->midiCtr1Ctrl(), v);

			break;
							
		case PANEL01_POT_CTRL2 	: 
							
			s->setPot(1, v); 
			dsp->update(DSPP_HV_CTR2, v << 16); 
			if(global->midiCtrlTransmit()) midi->newCtrlChangeEvent(global->midiCtr2Ctrl(), v);

			break;
							
		case PANEL01_POT_CTRL3 	: 
							
			s->setPot(2, v); 
			dsp->update(DSPP_HV_CTR3, v << 16); 
			if(global->midiCtrlTransmit()) midi->newCtrlChangeEvent(global->midiCtr3Ctrl(), v);

			break;
	}
}

void EntitySynth::PanelEncoderMod() {

	em += mBuffer[2] << 4;
	em = em < 0 ? 0 : em > 16383 ? 16383 : em;
	
	dsp->update(DSPP_HV_CTR1, em << 9); 
	if(global->midiCtrlTransmit()) midi->newCtrl14bEvent(global->midiCtr1Ctrl(), em);
}

void EntitySynth::ParameterSelect() { // User selects new parameter from front panel

	paraForEdit = paras->getDspPara((unsigned char)mBuffer[1]);
	sourceForEdit = (signed char)mBuffer[2]; 
	
	// This is for voice allocation paras "Polyphony" and "Stack"
	if(!paraForEdit->isDspParameter()) return;  
	
	// Put dsp para index in bits 0..7, source (0..n) in bits 8..15
	dspParaIndex = (unsigned char)mBuffer[1] | ((int)(sourceForEdit+1) << 8);

	dsp->update(DSPP_HV_PARAMETER, dspParaIndex);
	
	if(paraForEdit->isContinuous(sourceForEdit))
			
		dsp->command(DSPP_HV_SET_SMOOTH_EDIT);
	else	
		dsp->command(DSPP_HV_SET_DIRECT_EDIT);
}

void EntitySynth::ParameterModify() { // User changes parameter value from front panel
	
	EntitySound *s = pendingSound ? pendingSound : activeSound;
	int newValue = s->getTranslatedValue(paraForEdit, sourceForEdit);
	
	// This is for voice allocation paras "Polyphony" and "Stack"
	
	if(!paraForEdit->isDspParameter()) { 

		/* NB, this never gets called during stack edit mode,
		 * where it would be a mistake to reset because that would
		 * clear the special voice assignment settings used during stack edit */

		/* Reset voice assignment on next note-on */
		assign->resetVoiceAssign();
	}

	// This is for DSP paras (all other parameters)
	
	else { 
		updateDSPParaIndex(paraForEdit, sourceForEdit);
		dsp->update(DSPP_HV_PARAMETER_DATA, newValue);	
	}

	// This is for both
	
	if(global->midiEditTransmit()) {

		int value = 8192 + s->getValue(paraForEdit, sourceForEdit);
		midi->newNonRegParamEvent(paraForEdit->getParaNumber(), sourceForEdit+1, value);
	}
}	

void EntitySynth::MidiNRPNHiResUpdate() { // NRPN data is same as user value (plus offset)
	
	char midiSource = mBuffer[1] - 1;	// -1..n
	char midiParaIndex = mBuffer[2];

	Parameter *midiPara = paras->getParaByIndex(midiParaIndex);
	int  midiValue = ((mBuffer[3] << 7) + mBuffer[4]) - 8192;
	
	updateNRPN(midiPara, midiValue, midiSource);
}

void EntitySynth::MidiNRPNLoResUpdate() { // NRPN data needs translation according to parameter
	
	char midiSource = mBuffer[1] - 1;	// -1..n
	char midiParaIndex = mBuffer[2];
	
	Parameter *midiPara = paras->getParaByIndex(midiParaIndex);
	int midiValue = midiPara->translateNRPNPotValue(mBuffer[3], midiSource);
	
	updateNRPN(midiPara, midiValue, midiSource);
}

void EntitySynth::MidiNRPNIncrement() { // NRPN increment user value
	
	char midiSource = mBuffer[1] - 1;	// -1..n
	char midiParaIndex = mBuffer[2];
	
	Parameter *midiPara = paras->getParaByIndex(midiParaIndex);
	
	incrementNRPN(midiPara, 1, midiSource);
}

void EntitySynth::MidiNRPNDecrement() { // NRPN decrement user value
	
	char midiSource = mBuffer[1] - 1;	// -1..n
	char midiParaIndex = mBuffer[2];
	
	Parameter *midiPara = paras->getParaByIndex(midiParaIndex);
	
	incrementNRPN(midiPara, -1, midiSource);
}

void EntitySynth::incrementNRPN(Parameter *midiPara, int inc, int midiSource) {

	EntitySound *s = pendingSound ? pendingSound : activeSound;
	
	s->incValue(midiPara, inc, midiSource); // update sound object user value

	/* If NRPN matches currently displayed edited parameter, then update displayed value */
	if(midiPara == paraForEdit && midiSource == sourceForEdit) sendMessage(ctrq, DISPLAY_MIDI_EDIT);

	/* This is for voice allocation paras "Polyphony" and "Stack" */
	if(!midiPara->isDspParameter()) { 
		
		/* Reset voice assignment on next note-on */
		assign->resetVoiceAssign();
		
		return;
	}
	
	else {
		updateDSPParaIndex(midiPara, midiSource);

		int newValue = s->getTranslatedValue(midiPara, midiSource);
		dsp->update(DSPP_HV_PARAMETER_DATA, newValue);
	}
}

void EntitySynth::updateNRPN(Parameter *midiPara, int midiValue, int midiSource) {

	EntitySound *s = pendingSound ? pendingSound : activeSound;

	s->changeValue(midiPara, midiValue, midiSource); // update sound object user value

	/* If NRPN matches currently displayed edited parameter, then update displayed value */
	if(midiPara == paraForEdit && midiSource == sourceForEdit) sendMessage(ctrq, DISPLAY_MIDI_EDIT);

	/* This is for voice allocation paras "Polyphony" and "Stack" */
	if(!midiPara->isDspParameter()) { 
		
		/* Reset voice assignment on next note-on */
		assign->resetVoiceAssign();
		
		return;
	}
	
	else {
		updateDSPParaIndex(midiPara, midiSource);

		int newValue = s->getTranslatedValue(midiPara, midiSource);
		dsp->update(DSPP_HV_PARAMETER_DATA, newValue);
	}
}

void EntitySynth::updateDSPParaIndex(Parameter *newPara, int newSource) {

	if(!newPara->isDspParameter()) TRACE("ERROR - trying to send non-DSP para to DSP\n");
	
	int i = newPara->getDspParaIndex() | ((int)(newSource+1) << 8);
	
	if(dspParaIndex != i) {

		dspParaIndex = i;
	
		dsp->update(DSPP_HV_PARAMETER, dspParaIndex);
	
		if(newPara->isContinuous(sourceForEdit))
			
			dsp->command(DSPP_HV_SET_SMOOTH_EDIT);
		else	
			dsp->command(DSPP_HV_SET_DIRECT_EDIT);
	}
}

void EntitySynth::MidiNoteOn() {

	note = mBuffer[1];
	int velocity = mBuffer[2];
	
	if(!pendingSound) {
		
		assign->noteOn(note, velocity);
	
	} else { 
		
		/* NB activeSound needs to be valid for kflw() */
		
		activeSound = pendingSound;
		pendingSound = 0;
		
		dsp->update(DSPP_HV_CTR1, activeSound->getPot(0) << 16);
		dsp->update(DSPP_HV_CTR2, activeSound->getPot(1) << 16);
		dsp->update(DSPP_HV_CTR3, activeSound->getPot(2) << 16);

		em = activeSound->getPot(0) << 7;
		
		// A new sound is pending in the DSP, and on this new note it
		// will be copied to live locations in DSP memory. During this
		// transfer, no data should be written to the DSP, as it could
		// get lost. Furthermore, by avoiding communication during the
		// transfer, state relationships are simplified. This should 
		// cause a maximum delay of about 1.33 mS.
		
		dsp->inhibitDsp();
		dsp->command(DSPP_HV_ACTIVATE_NEW_SOUND);

		assign->noteOn(note, velocity, activeSound, Wait_for_dsp);
	}
}
	
void EntitySynth::MidiNoteOff() { assign->noteOff(mBuffer[1]); }

void EntitySynth::MidiSustain() { 
	
	if(mBuffer[1] > 63) assign->sustain(true); 
	else assign->sustain(false);
}

void EntitySynth::MidiController() {
	
	int v = (mBuffer[2] << 7) | mBuffer[3]; // Original unsigned 14bit midi value

	// Source: CTR1 = 1, CTR2 = 2, CTR3 = 4, MMOD = 8
						
	char s = mBuffer[1]; // Modulation selection bits

	// The theory here is that one controller could be mapped to several 
	// matrix sources. This would not be much use if set deliberately, 
	// but it's also possible to encounter this situation dynamically 
	// (while editing the rtc assignment), and also when a ctr is not 
	// in use, a double assignment might occur without the user noticing. 
	// The system should work appropriately under all such conditions.
						
	EntitySound *snd = pendingSound ? pendingSound : activeSound;

	if(s&0x01) { dsp->update(DSPP_HV_CTR1, v << 9); snd->setPot(0, v >> 7); em = v; }
	if(s&0x02) { dsp->update(DSPP_HV_CTR2, v << 9); snd->setPot(1, v >> 7); }
	if(s&0x04) { dsp->update(DSPP_HV_CTR3, v << 9); snd->setPot(2, v >> 7); }
		
	if(s&0x08) dsp->update(DSPP_HV_MMOD, v << 9);
}

void EntitySynth::MidiPitch() {

	int midiPitchWheel = ((mBuffer[1] << 7) | mBuffer[2]);
	dsp->update(DSPP_HV_PITCH_SHIFT, pitchShift(midiPitchWheel) << 8);
}	

void EntitySynth::MidiVolume() {

	volume = (mBuffer[1] << 7) | mBuffer[2];
	int a = (volume * expression) >> 14;
	dsp->update(DSPP_HV_MASTER_VOL, a << 9);
}	
						
void EntitySynth::MidiPan() {

	pan = ((mBuffer[1] << 7) | mBuffer[2]) - 8192;
	dsp->update(DSPP_HV_MIDI_PAN, pan << 10);
}	
						
void EntitySynth::MidiExpression() {

	expression = (mBuffer[1] << 7) | mBuffer[2];
	int a = (volume * expression) >> 14;
	dsp->update(DSPP_HV_MASTER_VOL, a << 9);
}	
						
void EntitySynth::MidiReset() {

	assign->allNotesOff();
	
	pan = 0;
	volume = 16383;
	expression = 16383;
}

void EntitySynth::ProgramChange() {

	// This event is generated by sound mode when the
	// user selects a new sound, purely so that this
	// information can be transmitted over midi.
	
	if(!global->midiProgTransmit()) return;

	// Sound bank and user sound number are sent

	midi->newProgChangeEvent(mBuffer[1], mBuffer[2]);  
}
		
void EntitySynth::GlobalTune() {

	// Derivation: 
	// k = 440 = 2^(v/12).440 - f
	// where v is compensation for freq f in semitones,
	// 2^(v/12) = 1 + f / k
	// v = 12.log2(1+f/k)
	// or v = 4800.log2(1+f/k) in  cent units
	
	double F = (double) global->masterTuning() / 100.0; // Freq correction in Hz
	double V = 4800.0 * log2(1.0 + F / 440.0);	    // Pitch correction result

	int v = (int)(V * 256.0);	// DSP prefers pitch at 24 bits
	
	dsp->update(DSPP_HV_MASTER_TUNE, v);
}	
	
void EntitySynth::NoteEvent() {

	/* Note events for transmission over midi */
	
	int n = mBuffer[1];
	int v = mBuffer[2];

	if(v) {

		midi->newKeyOnEvent(n, v);

	} else {

		midi->newKeyOffEvent(n);
	
		/* Useful for clearing hanging notes */
		midi->newAllNotesOffEvent();
	}	
}

void EntitySynth::AllNotesOff() {

	midi->newAllNotesOffEvent();
}

void EntitySynth::MidiAllNotesOff() {

	assign->allNotesOff();
}

void EntitySynth::MidiAllSoundOff() {

	assign->allSoundOff();
}

void EntitySynth::NewSoundReady() {
	
	loadNewSound(); 
}			

void EntitySynth::ReadDSPMIPS() {

    	int x = dsp->readTotalMips();
	if(x < 0 ) x = 0;					
	mBuffer[0] = AUTO_TOTAL_MIPS;
	mBuffer[1] = (x>>16) & 0xff;
	mBuffer[2] = (x>>8) & 0xff;
	mBuffer[3] = x & 0xff;

	sendMessage(ctrq, mBuffer, 4);

	x = dsp->readFragMips();
	if(x < 0 ) x = 0;					
	mBuffer[0] = AUTO_FRAG_MIPS;
	mBuffer[1] = (x>>16) & 0xff;
	mBuffer[2] = (x>>8) & 0xff;
	mBuffer[3] = x & 0xff;
	
	sendMessage(ctrq, mBuffer, 4);
}

void EntitySynth::ReadDSPDebugData() {
	
	int x = dsp->readDebugData();

       mBuffer[0] = AUTO_DEBUG_DATA;
       mBuffer[1] = (x>>16) & 0xff;
       mBuffer[2] = (x>>8) & 0xff;
       mBuffer[3] = x & 0xff;
       
       sendMessage(ctrq, mBuffer, 4);
}

void EntitySynth::ResetDSPDebugAddress() {

	dsp->command(DSPP_HV_RESET_DEBUG_ADDRESS);
}

void EntitySynth::NullEvent() {

	TRACE("ERROR: Null message vector %i in EntitySynth.cpp\n", mBuffer[0]);
}

void EntitySynth::ClearDSPLevels() {

	// Reading levels resets them to zero
	
	dsp->readMaxInLevels();
	dsp->readMaxOutLevels();
}

void EntitySynth::ReadInputLevels() {

	int levels = dsp->readMaxInLevels();

	int left = levels & 0xFFF;
	int right = (levels >> 12) & 0xFFF;

	mBuffer[0] = AUTO_INPUT_LEVELS;
	mBuffer[1] = meterLevel.convert(left);
	mBuffer[2] = meterLevel.convert(right);
	
	sendMessage(ctrq, mBuffer, 3);
}

void EntitySynth::ReadOutputLevels() {

	int levels = dsp->readMaxOutLevels();

	int left = levels & 0xFFF;
	int right = (levels >> 12) & 0xFFF;

	mBuffer[0] = AUTO_OUTPUT_LEVELS;
	mBuffer[1] = meterLevel.convert(left);
	mBuffer[2] = meterLevel.convert(right);

	sendMessage(ctrq, mBuffer, 3);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// 		       	Class for calculating meter levels
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

char MeterLevel::convert(int peak) { return lut[peak & 0xFFF]; }

MeterLevel::MeterLevel() {

#define FAST_METER_LEVEL_INIT

#ifdef FAST_METER_LEVEL_INIT

	const int lutValueCount[] = {	// These value counts obtained from slow init

	   // 	0, 1, 2, 3, 4, 5, 6,  7,  8,  9,  10,  11,  12,  13,  14,  15,  16
		2, 1, 2, 3, 6, 9, 16, 25, 44, 73, 124, 207, 349, 586, 986, 1659, 4
	};

	const int nDiffValues = 17;
	
	for(int x = 0, y = 0; y < nDiffValues; y++) {

		for(int z = 0; z < lutValueCount[y]; z++) {

			lut[x++] = y;
		}
	}

#else 	// Slow method: 4096 logarithms take too long (seconds) for normal initialisation
	// This code is included here to show derivation of fast method.
	
	const float max = 4096.0;		// Numerical range of signals
	const float segments = 16.0;		// Number of segments on display

	float meterSaturation = segments;	// To make top segment the saturation indicator
	float peakThreshold = max - 5.0;	// Any values above this are displayed as saturated
	float logPT = log10(peakThreshold);	// The maximum value coming out of the log function

	float scale = meterSaturation / logPT;	// Scaling factor for meter after log function

	for(int x = 0; x < 4096; x++) { lut[x] = (char) (scale * log10(x)); }

	for(int n = 0; n <= 16; n++) {		// Generate values for lutValueCount[]

		int count = 0; for(int x = 0; x < 4096; x++) if(lut[x] == n) count++;
		TRACE("lut[..] = %i:	%i\n", n, count); 
	}
#endif

}	

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// 				    NON-VECTORS
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

int EntitySynth::kflw(int midiNoteNumber) {

	// KFLW is derived from the note. When a new note comes in, 
	// its pitch in cents is calculated here.

	// Each sound has its own tuning data, which can be programmed using the 
	// MIDI Tuning Standard. This is stored with the sound. No Interface is 
	// provided for tuning or scale selection.

	// Calculate value offset from middle C in DSP pitch units (ie 400 steps-per-semitone)
	
	int kflw = 0;

	if(activeSound) kflw += activeSound->gettuning(midiNoteNumber);	// access tuning data for sound

	// The pitch units are selected for an LUT size of 65536 words, so limit the range:

	return kflw > 32767 ? 32767 : kflw < -32768 ? -32768 : kflw;
}

int EntitySynth::pitchShift(int midiPitchWheel) {

	// Pitch data format for the dsp is  cent units offset from middle C (note 60)
	
	int PITCH_RESOLUTION = 400; 	// Tuning steps-per-semitone ( cent resolution)

	int pitchShiftDsp = (midiPitchWheel - 8192) * global->midiPitch();	// in 8192 units-per-semitone

	pitchShiftDsp = (pitchShiftDsp * PITCH_RESOLUTION) >> 13;  // (divide by 8192)

	// The pitch units are selected for an LUT size of 65536 words, so limit the range:

	return pitchShiftDsp > 32767 ? 32767 : pitchShiftDsp < -32768 ? -32768 : pitchShiftDsp;
}

void EntitySynth::loadNewSound() {

	/* For full details on the operations below see EntitySoundManager.txt */
	/* This has been updated to include transfer of voice assignment context data */

	EntitySound *s = 0;
	bool moreSounds = true;

	if(pendingSound) pendingSound->makeAvailable();
	pendingSound = 0;
	
	while(moreSounds) { 
		
		/* Read sound objects placed on the new sound stack by the sound manager */

		s = getSoundFromStack(); // (read message from SSTQ without blocking)
		
		/* Is there a newer sound to load? If so, make this older one available 
		 * for re-use in the sound pool. This reciprocates sound manager actions. */
		
		if(s && (moreSounds = moreSoundsOnStack())) s->makeAvailable();

		else pendingSound = s; // This is the latest sound object on the stack
	}

	/* It is perfectly feasible for there to be no sounds in the stack after the 
	 * NEW_SOUND_READY message was received on SYNQ. This is because SYNQ messages 
	 * can be significantly delayed, whereas the stack will always have the latest
	 * sound as the last entry. When scanning, it may be common for this function
	 * to be called with an empty sound stack (ie no messages on SSTQ). */

	#ifdef DEBUG_SYNTH
	TRACE("Initiating sound transfer to dsp\n");
	#endif

	if(!pendingSound) return; // Do nothing if stack was empty; sound transfer already done

	if(activeSound) activeSound->makeAvailable();
	activeSound = 0;

	dsp->update(pendingSound);		

	/* Load default values for CTR1-3 (preferred over using pot values stored with sound)
	 * These are initialised here, so changes to CTR1-3 before first note can affect the 
	 * pending sound (at present, such changes also affect the end of the current sound) */
	   
	pendingSound->setPot(0, 64);
	pendingSound->setPot(1, 64);
	pendingSound->setPot(2, 64);

	/* The following code causes a new sound to be made current immediately
	 * (in Entity polyphonic, this mutes all voices until a new note is played) */
	
	#ifndef NEW_SOUND_WAITS_FOR_NEW_NOTE

	activeSound = pendingSound;
	pendingSound = 0;
							
	dsp->inhibitDsp();
	dsp->command(DSPP_HV_ACTIVATE_NEW_SOUND);
	dsp->waitForDsp(); // Wait for internal DSP transfer (should be max 1.33 mS)

	#endif
}

EntitySound *EntitySynth::getSoundFromStack() {

	EntitySound *sound;
	rtems_unsigned32 nBytes;

	status = rtems_message_queue_receive(sstq, &sound, &nBytes, RTEMS_NO_WAIT, RTEMS_NO_TIMEOUT);

	switch(status) {

		case RTEMS_SUCCESSFUL :		if(nBytes == 4) return sound; 
						else TRACE("ERROR: SSTQ message size\n");
						
		case RTEMS_UNSATISFIED :	return 0; // Queue empty						

		default : 			TRACE("ERROR: Some kind of SSTQ problem\n");
	}

	return 0;
}

bool EntitySynth::moreSoundsOnStack() {

	rtems_unsigned32 count;
	rtems_message_queue_get_number_pending(sstq, &count);
	return (bool) count;
}

