/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityVoiceAssigner.cpp						*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005                                      *
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/
  
#include "EntityVoiceAssigner.h"

// Comment out these lines when no longer needed

//#define DEBUG_VOICE_ASSIGNER
//#define DEBUG_VOICE_ASSIGNER_DETAIL

// Some debug code for tracking voice assignment states

#ifdef  DEBUG_VOICE_ASSIGNER_DETAIL
#define TRACE_VOICE_ASSIGNER trace
#else
#define	TRACE_VOICE_ASSIGNER notrace
#endif

// EntityVoiceAssigner translates note events into voice events

void EntityVoiceAssigner::noteOn(int note, int velocity) { 

	if(!sound) return;
	
	if(contextPending) {
		
		allNotesOff();
		reset(sound);
		newNote(note, velocity, Note_on);
		return;
	}
		
	newNote(note, velocity, Note_on);
	
	TRACE_VOICE_ASSIGNER();
}

void EntityVoiceAssigner::noteShift(int note) { 
	
	/* No velocity, since note shift doesn't restart an already sounding note */
	newNote(note, 0, Note_shift);
	
	TRACE_VOICE_ASSIGNER();
}

void EntityVoiceAssigner::noteOn(int note, int velocity, EntitySound *sound_, DspWaitAction dspAction) { 

	reset(sound_);
	
	if(dspAction == Wait_for_dsp) {
		
		/* Wait for internal DSP transfer (max ~3 mS). Hopefully this
		 * will have occurred in parallel with reset(sound_) */

		dsp->waitForDsp(); 	
	}
	
	newNote(note, velocity, Note_on);
	
	TRACE_VOICE_ASSIGNER();
}

void EntityVoiceAssigner::newNote(int note, int velocity, NoteAction action) { 

	if(!sound) { 
		
		TRACE("ERROR: No sound defined in EntityVoiceAssigner::noteOn(..)\n");
		return;
	}
	
	if(context->echo) {

		/* If editing the stack, the editor needs to know what notes are played */
		
		mBuffer[0] = (char) MIDI_NOTE_ON;
		mBuffer[1] = (char) note;
		mBuffer[2] = (char) velocity;

		sendMessage(ctrq, mBuffer, 3);
	}

	int nStackedPerNote = context->nStacked;

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Assign voices to note if note doesn't already have any
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	NoteListItem *n = notes[note];

	/* Get pointer to first voice in note's voice stack - the validity of this 
	 * pointer is indicated by the number of voices assigned to the note */

	VoiceListItem *firstVoice = n->getVoices();
	
	/* If note has no voices assigned to it then assign voices */
	if(!n->getnVoices()) { 

		/* If the sound's portamento is above POLY_PORTAMENTO_THRESHOLD then 
		 * assign voices from the *top* of voices-off. Otherwise, assign from the 
		 * bottom of voices-off. If voices-off is empty, then assign from
		 * the bottom of voices-on, in either case. This is an implementation 
		 * of polyphonic portamento. */

		/* Note that, when playing monophonically, the contents of voices-off
		 * can only ever be one note's voice stack, or empty, so it makes
		 * no difference which end is selected for voice assignment */
		
		/* If voices-off has some voices in it */
		if((firstVoice = topOfVoicesOff->getDown()) != bottomOfVoicesOff) {

			/* firstVoice already points to the voices at the top of voices-off, but
			 * if not using polyphonic portamento, assign from bottom of voices-off */
			
			if(sound->getValue(portamento) < POLY_PORTAMENTO_THRESHOLD) { 
			
				firstVoice = bottomOfVoicesOff;
			
				/* Scan up from bottom of voices off, according to number of voices stacked */
				for(int v = 0; v < nStackedPerNote; v++) firstVoice = firstVoice->getUp();
			}

		} else {
			
			/* Voices-off was empty, so we need to find voices from the bottom of voices-on */
		
			firstVoice = bottomOfVoicesOn;
			
			/* Scan up from bottom of voices on, according to number of voices stacked */
			for(int v = 0; v < nStackedPerNote; v++) firstVoice = firstVoice->getUp();
		}

		/* firstVoice now contains the first of a linked list of voices, of length according to
		 * the number stacked, that need to be de-assigned from their current note, and re-assigned
		 * to the new note */
	}

	/* De-assign voices from their previous note (its voice pointer doesn't need to be correct if nVoices is zero) */

	int oldnote = firstVoice->getNote(); // if this value is out of range, then don't try to access that note!
	if(oldnote >= 0 && oldnote < N_MIDI_NOTES) notes[oldnote]->setnVoices(0); 

	if(action != Note_shift) {

		/* In both cases below, the DSP will force staccato play if the
		 * note was released and portamento is effectively zero */
	
		/* If restarting voices on same note, don't reset envelopes to zero */
		if(oldnote == note) action = Note_on_smooth;
	
		/* If monophonic, don't reset envelopes to zero */
		if(!context->polyphonic) action = Note_on_smooth;
	}
	
	/* Assign voices to new note */
	n->setVoices(nStackedPerNote, firstVoice);
		
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Update voices assigned to the note
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	/* Note has voices assigned to it, so restart these voices, move them to
	 * the top of voices-on, and move the note itself to top of notes-on */
	
	/* Identify first voice assigned to this note */
	VoiceListItem *voice = firstVoice = n->getVoices(); 	
	
	/* Restart voices assigned to this note */

	for(int v = 0; v < nStackedPerNote; v++) {

		int i = voice->getVoiceIndex();

		if(i != -1) {	// if voice is rendered on this machine

			int stackNote = note + context->stacklist[v];		// Transposed note
			
			if(stackNote < 0 || stackNote > 127) { 			// If note is off the scale...

				dsp->voiceMute(i);				// ...then mute it
			}

			else {							// If note is on the scale...

				int pitch = sound->gettuning(stackNote);	// Look up pitch in sound's tuning
				
				/* And restart the voice */

				dsp->voiceAssign(voice->getVDIF(), voice->getSTDF(), i);

				#ifdef DEBUG_VOICE_ASSIGNER
				TRACE("Voice #%d \tON\tnote = %d\tvelocity = %d\n", i, stackNote, velocity);
				#endif
				
				switch(action) {

					case Note_on :
						
						/* Reset voice envelopes and apply velocity */
						dsp->voiceOn(pitch, velocity);
						break;
					
					case Note_on_smooth :
						
						/* Reset voice envelopes without clearing their levels to zero 
						 * (attack starts from current envelope level) */
						dsp->voiceOnSmooth(pitch, velocity);
						break;

					case Note_shift :
						
						/* Shift voice pitch, leaving velocity & envelopes unchanged */
						dsp->voiceShift(pitch); 
						break;
				}
			}
		}

		/* Update voice's internal note reference - this is very important: without a voice
		 * knowing which note it is assigned to, it cannot be de-assigned from it in the
		 * first step of the process (see first part of this method) where voices are
		 * re-assigned to a different note. */ 

		voice->setNote(note);		

		/* Scan to next voice */
		voice = voice->getDown();	
	}

	/* Move note to top of notes-on and set state indicating it is physically on */
	notes[note]->moveBelow(topOfNotesOn);
	notes[note]->setOn(true);

	/* Move the voice stack to the top of voices-on */
	firstVoice->moveVoiceStackBelow(topOfVoicesOn, voice->getUp());
}

void EntityVoiceAssigner::noteOff(int note) { 

	if(!sound) return;
	
	/* If sustain pedal is on, then no linked list structure change,
	 * but set the sustain bit for the note, indicating it's no
	 * longer actually physically played */
	
	if(context->sustain) {

		notes[note]->setOn(false);	// Indicate key is physically off
		return;				// But don't do anything else
	}
		
	/* If note wasn't on (midi mistake), then ignore this event */
	
	if(!notes[note]->getOn()) return;
	
	/* Note was on, so move it to top of notes-off */
	
	NoteListItem *n = notes[note];		// Find note's list entry
	
	n->setOn(false);			// Note is physically off
	n->moveBelow(topOfNotesOff);		// Move note to notes-off

	/* If no voices assigned to the note, exit */
	
	if(!n->getnVoices()) {
		
		TRACE_VOICE_ASSIGNER();
		return;
	}
	
	/* If polyphonic, put note's voices into release phase */
	
	if(context->polyphonic) {
	
		/* Put the voices assigned to the note into release */
		voicesOff(n);
		
		/* For polyphonic portamento, de-assign voices from note, so when the note
		 * is played next time, voices from the portamento start note are used */
		
		if(sound->getValue(portamento) > POLY_PORTAMENTO_THRESHOLD) {
			
			n->setnVoices(0);
			
			/* You also have to make sure that the voices don't think
			 * they're still assigned to that note, else when they get
			 * re-used, the current note, with its new set of voices, 
			 * could have its number of voices zeroed without the voices
			 * being re-used. */
		
			VoiceListItem *voice = n->getVoices();
		
			for(int v = 0; v < context->nStacked; v++) {

				voice->setNote(-1);
				voice = voice->getDown();
			}
		}
	
		TRACE_VOICE_ASSIGNER();

		return;
	}
	
	/* Monophonic: so revert to any note that's still on */

	NoteListItem *mostRecentNote = topOfNotesOn->getDown();	

	if(mostRecentNote == bottomOfNotesOn) {

		/* No voices remain in notes-on on, so put voices 
		 * (if any) of released note into release phase */
	
		voicesOff(n);
	
		TRACE_VOICE_ASSIGNER();
	
		return;
	}
	
	/* Revert to note still held */
		
	noteShift(mostRecentNote->getNoteIndex());
	
	TRACE_VOICE_ASSIGNER();
}

bool EntityVoiceAssigner::sustain(bool s) { 

	if(!sound) return s;
	
	/* If sustain on, just set sustain flag */
	if((play.sustain = s)) return s;	

	/* If no notes on, do nothing */
	if(topOfNotesOn->getDown() == bottomOfNotesOn) return s;	

	if(context->polyphonic) sustainReleasePolyphonic();
	else sustainReleaseMonophonic();

	return s;
}

void EntityVoiceAssigner::sustainReleasePolyphonic() {

	/* Go through notes-on and check for any physically released keys 
	 * that are sustained, releasing voices as we go */

	NoteListItem *n = bottomOfNotesOn->getUp();
	
	while(n != topOfNotesOn) {

		NoteListItem *nn = n->getUp();		// Next note up

		if(!n->getOn()) {			// If key is physically off
		
			voicesOff(n);			// Put note's voices in release phase
			n->moveBelow(topOfNotesOff);	// Move note to notes-off
		}
		
		n = nn;					// Current note = next note
	}
}

void EntityVoiceAssigner::sustainReleaseMonophonic() {

	/* First go through notes-on and move any released keys
	 * to notes-off, without updating voices. */
	
	/* Make note of most recent note played */
	NoteListItem *mostRecentNote = topOfNotesOn->getDown();
	
	NoteListItem *n = bottomOfNotesOn->getUp();

	while(n != topOfNotesOn) {

		NoteListItem *nn = n->getUp();		// Next note up

		if(!n->getOn()) {			// If key is physically off
		
			n->moveBelow(topOfNotesOff);	// Move note to notes-off
		}
		
		n = nn;					// Current note = next note
	}

	/* If no notes left in notes-on, then release voices assigned to most recent note */

	if(topOfNotesOn->getDown() == bottomOfNotesOn) {

		voicesOff(mostRecentNote);
		return;
	}
	
	/* Monophonic revert-to-most-recent-remaining */
	
	/* If most recent note has not changed, do nothing */ 
	if(mostRecentNote == topOfNotesOn->getDown()) return;

	/* Otherwise, shift pitch to that of new most recent note */
	mostRecentNote = topOfNotesOn->getDown();
	noteShift(mostRecentNote->getNoteIndex());
}

void EntityVoiceAssigner::allSoundOff() { 

	/* Mute all voices rendered on this machine */
	for(int v = 0; v < N_VOICES_PER_MACHINE; v++) dsp->voiceMute(v);
	
	/* Release all voices, so allocation works as expected
	 * when the next note is played */
	sustain(false);
	allNotesOff();
} 

void EntityVoiceAssigner::allNotesOff() { 

	if(!sound) return;
	
	if(!context->sustain) { 			// if sustain pedal isn't on

		NoteListItem *n = bottomOfNotesOn->getUp();

		while(n != topOfNotesOn) {

			voicesOff(n);			// Put note's voices in release phase
			n->setOn(false);		// Note is physically off
			n->moveBelow(topOfNotesOff);	// Move note to notes-off

			n = bottomOfNotesOn->getUp();
		}

		return;
	}

	/* If sustain pedal is on, then no linked list structure change,
	 * but set the sustain bit for each note */

	NoteListItem *n = bottomOfNotesOn->getUp();

	while(n != topOfNotesOn) {

		n->setOn(false);	// Indicate key is physically off
		n = n->getUp();		// Next note up
	}
}

void EntityVoiceAssigner::voicesOff(NoteListItem *n) {

	/* This places any voices assigned to the note into release phase. There can be zero, 
	 * one or several voices. The voices are then placed in the voices-off linked list */
	
	int nv = n->getnVoices();		// Number of voices assigned to this note
	
	/* Don't try to release voices that don't belong to the note */
	if(!nv) return; 

	VoiceListItem *firstVoice;
	VoiceListItem *voice;
	
	voice = firstVoice = n->getVoices();	// First voice in linked list of n voices assigned to this note

	for(int v = 0; v < nv; v++) {

		int vi = voice->getVoiceIndex();
		
		#ifdef DEBUG_VOICE_ASSIGNER
		TRACE("Voice #%d \tOFF\n", vi);
		#endif
		
		/* if voice is rendered on this machine, put its envelopes into release */ 
		if(vi != -1) dsp->voiceOff(vi);
		
		voice = voice->getDown();
	}

	firstVoice->moveVoiceStackBelow(topOfVoicesOff, voice->getUp());
}

void EntityVoiceAssigner::reset(EntitySound* sound_) {

	if(contextPending) {
	
		/* This is the only place where context should be updated,
		 * to guarantee that this is synchronous with reset */
		
		context = contextPending;
		contextPending = 0;
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
	// Update current sound pointer and set play context by reading sound parameters
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
	//
	//	struct Context {
	//
	//		bool	polyphonic;	// True if polyphonic voice assignment required
	//		int	*stacklist;	// List of note transpositions for stacked voices
	//		int	nStacked;	// Minimum value is 1
	//		bool	echo;		// Echo note-on events to editor
	//		bool	sustain;	// Updated from midi (or always on in stackEdit)
	//	};
	//	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

	sound = sound_;

	/* Update play.polyphonic, .stacklist[] and .nStacked from "polyphonic" and "stack" */
	
	play.stacklist[0] = 0;	// First voice in a stack is always untransposed
	play.nStacked = 1;	// Default one voice in stack (ie, no voice stacking)

	if(sound->getValue(stack)) {	// If stack is "ON"

		for(int s = 1; s < MAX_NOTE_STACK; s++) { 

			int transpose = sound->getValue(stack, s-1);	

			if(transpose != CLEAR_STACK_NOTE) {
			
				play.nStacked++;
				play.stacklist[s] = transpose;
			}
		}
	}

	play.polyphonic = (bool) sound->getValue(polyphonic);

	/* Update from global parameters */
	
	nMachines = global->midiCascadeGetnMachines();
	machineIndex = global->midiCascadeGetMachineIndex();

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
	// On entry it is assumed that voice rendering has been updated for release or mute
	// so that it is unnecessary to individually examine each voice to check 
	// whether this needs to be done.
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

	// Move any remaining notes on to notes-off

	NoteListItem *n;
	
	while((n = bottomOfNotesOn->getUp()) != topOfNotesOn) {

		#ifdef DEBUG_VOICE_ASSIGNER
		TRACE("Moving note from notes-on to notes-off\n");
		#endif
		
		n->moveBelow(topOfNotesOff);	// Move to notes-off
	}

	/* Clear any voices assigned to notes */
	for(int ni = 0; ni < N_MIDI_NOTES; ni++) notes[ni]->setnVoices(0);
	
	// The number of voices to be made available is a multiple of the number of voices that
	// will be stacked on each note. This avoids partial re-assignment of a note's voices.

	int nStackedPerNote = context->nStacked;

	int nVoicesAvailable = nMachines * N_VOICES_PER_MACHINE;
	int nVoicesRem = nVoicesAvailable % nStackedPerNote;
	int nVoicesUsed = nVoicesAvailable - nVoicesRem;

	if(!context->polyphonic) nVoicesUsed = nStackedPerNote;
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Calculate voice difference (VDIF) and stack difference (STDF) for each of the available voices
	// VDIF is different for each note, but each voice assigned to the same note has the same VDIF
	// value. STDF is different for each voice stacked in a note, but the same STDF values are
	// used in other notes.
	//
	// This is worth explaining in some detail. VDIF (possibly to be renamed NDIF) is a value in the
	// matrix that is different for each note played, but the same for each voice stacked in that note.
	// The idea here is to provide a value to modulate a parameter such as pan, thereby achieving an
	// effect such as spreading a chord across the stereo field, varying LFO rates for each note etc.
	// STDF does a similar thing for voices within a note, so that when voices are stacked to the same
	// note, they can be detuned, spread, etc (although this implementation is slightly flawed, because 
	// combined pan and detune has the effect of making the left flat and the right sharp, which 
	// dedicated stack detune processing can overcome, to some degree, when three or more voices 
	// are stacked. This is a limitation of the current design).
	//
	// If there is six note polyphony, there need to be six different values for VDIF spread across
	// the range 0..127. These would be: 0, 25, 51, 76, 102, 127. For monophonic playing, the value
	// of VDIF is 64. The same set of relationships is used for allocating STDF values within a note.
	// For example, with no stacking, STDF should be 64. With two voices per note, STDF = 0, 127. 
	//
	// Below, we calculate an array of values for VDIF and SDIF, so when we come to initialise the
	// linked list of voices, the values are just copied across into the appropriate voice list
	// item. Provided that voices are always reassigned in contiguous groups equal to the stack 
	// size, appropriate VDIF and STDF values will be available while playing without the need
	// for recalculation.
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	// Calculate note polyphony
	
	int nNotesPolyphony = nVoicesUsed / nStackedPerNote;

	#ifdef DEBUG_VOICE_ASSIGNER
	TRACE("\n---------------------------------\n");
	TRACE("nStacked = %d\n", nStackedPerNote);
	TRACE("nVoicesAvailable = %d\n", nVoicesAvailable);
	TRACE("nVoicesRem = %d\n", nVoicesRem);
	TRACE("nVoicesUsed = %d\n", nVoicesUsed);
	TRACE("nNotesPolyphony = %d\n", nNotesPolyphony);
	#endif

	// Enforce monophonic playing style (unlimited revert-to-notes-on)
	// if stacking results in single note polyphony (optional)

	if(nNotesPolyphony == 1) context->polyphonic = false;

	// Precalculate arrays of VDIF and STDF values to be assigned to voices
	
	signed char *VDIFptr;
	signed char *STDFptr;
	
	int VDIF = 0, STDF;	// Value that is incremented to calculate array below

	int STDFbase = 0; 	// Base value for each set of STDF values in a note's voices
	int VDIFincr = 0; 	// Amount to increment VDIF for each polyphonic note
	int STDFincr = 0;	// Amount to increment STDF for each stacked voice in a note

	// A shift of 16 bits is used to facilitate accurate accumulation during subsequent
	// calculations. The actual values supplied to the DSP during voice assignment are
	// in the range 0..127
	
	if(nNotesPolyphony > 1) { VDIF = 0; VDIFincr = ((128<<16)-1) / (nNotesPolyphony - 1); } 
	else 		  { VDIF = 64<<16; VDIFincr = 0; }

	if(nStackedPerNote > 1) { STDFbase = 0; STDFincr = ((128<<16)-1) / (nStackedPerNote - 1); } 
	else 		  	{ STDFbase = 64<<16; STDFincr = 0; }

	// So for no stacking and monophonic playing, we have VDIF & STDF = 64
	// For duophony, no stacking: VDIF = 0, 127, STDF = 64;
	// For monophony, triple stacking: VDIF = 0, STDF = 0, 64, 127; etc

	VDIFptr = VDIFlist;	// Use pointers for fast sequential list access
	STDFptr = STDFlist;
	
	#ifdef DEBUG_VOICE_ASSIGNER
	TRACE("VDIF/STDF array:\n");
 	#endif
	
	for(int n = 0; n < nNotesPolyphony; n++) {

		#ifdef DEBUG_VOICE_ASSIGNER
		TRACE("Note %d:\n", n);
		#endif	
		
		STDF = STDFbase; // Start new round of STDF values for next note
		
		for(int s = 0; s < nStackedPerNote; s++) {

			*VDIFptr++ = (signed char)(VDIF >> 16);
			*STDFptr++ = (signed char)(STDF >> 16);

			#ifdef DEBUG_VOICE_ASSIGNER
			TRACE("\t\tVDIF = %d\tSTDF = %d\n", VDIF>>16, STDF >> 16);
			#endif	
			
			STDF += STDFincr;
		}

		VDIF += VDIFincr;
	}
		
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Place available voices in voices-off
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	topOfVoicesOn->setDown(bottomOfVoicesOn); bottomOfVoicesOn->setUp(topOfVoicesOn); 
	topOfVoicesOff->setDown(bottomOfVoicesOff); bottomOfVoicesOff->setUp(topOfVoicesOff); 
	
	VDIFptr = VDIFlist;		// Array of pre-calculated VDIF & STDF values
	STDFptr = STDFlist;		// (using pointers for efficient access)

	for(int mi = 0, vi = 0; mi < nMachines; mi++) { // machine index, voice index

		// If these voices are rendered on this machine, give them a local voice index
			
		if(mi == machineIndex) { 

			for(signed char lvi = 0; lvi < N_VOICES_PER_MACHINE && vi < nVoicesUsed; lvi++, vi++) { 	

				voices[vi]->initialiseAbove(bottomOfVoicesOff, lvi, *VDIFptr++, *STDFptr++);
			}
		}

		// Otherwise, -1 indicates these voices are not rendered on this machine
		
		else {	for(signed char lvi = 0; lvi < N_VOICES_PER_MACHINE && vi < nVoicesUsed; lvi++, vi++) { 	

				voices[vi]->initialiseAbove(bottomOfVoicesOff, -1, *VDIFptr++, *STDFptr++);
			}
		}
	}

	#ifdef DEBUG_VOICE_ASSIGNER

	/* Let's have a look at the contents of voices-off */
	
	VoiceListItem *v = topOfVoicesOff->getDown();

	TRACE("Voices-off:\n");
	
	while(v != bottomOfVoicesOff) {

		TRACE("voiceIndex = %d\t", v->getVoiceIndex());
		TRACE("VDIF = %d\t", v->getVDIF());
		TRACE("STDF = %d\n", v->getSTDF());

		v = v->getDown();
	}

	TRACE("\n");	
	#endif
}

void EntityVoiceAssigner::resetVoiceAssign()	{ contextPending = &play; }
void EntityVoiceAssigner::startStackEdit() 	{ contextPending = &stackEdit; }
void EntityVoiceAssigner::endStackEdit()   	{ contextPending = &play; }

EntityVoiceAssigner::EntityVoiceAssigner(EntityGlobal* global_, ParaStructure* paras_, EntityDsp* dsp_) {

	dsp = dsp_;
	paras = paras_;
	global = global_;
	ctrq = getQueueId("CTRQ");

	sound = 0;
	
	for(int x = 0; x < MAX_VOICES; x++) {
		
		voices[x] = &voices_[x];
		voices[x]->setVoiceIndex(-99);
	}

	topOfVoicesOn = new VoiceListItem(-111); bottomOfVoicesOn = new VoiceListItem(-112);
	topOfVoicesOff = new VoiceListItem(-113); bottomOfVoicesOff = new VoiceListItem(-114);

	topOfVoicesOn->setDown(bottomOfVoicesOn); bottomOfVoicesOn->setUp(topOfVoicesOn); 
	topOfVoicesOff->setDown(bottomOfVoicesOff); bottomOfVoicesOff->setUp(topOfVoicesOff); 

	// (NB not initialising Voices-on or voices-off lists here, as these 
	// are dependent on the cascade parameters, which are variable)
	
	topOfNotesOn = new NoteListItem; bottomOfNotesOn = new NoteListItem;
	topOfNotesOff = new NoteListItem; bottomOfNotesOff = new NoteListItem;

	topOfNotesOn->setDown(bottomOfNotesOn); bottomOfNotesOn->setUp(topOfNotesOn); 
	topOfNotesOff->setDown(bottomOfNotesOff); bottomOfNotesOff->setUp(topOfNotesOff); 

	for(int x = 0; x < N_MIDI_NOTES; x++) {
		
		notes[x] = &notes_[x];
	
		notes[x]->insertAbove(bottomOfNotesOff);
		notes[x]->setOn(false);
		notes[x]->setnVoices(0);
		notes[x]->setNoteIndex(x);
	}

	// The values in play are variable, updated in response to events and sounds
	
	play.polyphonic = true;	
	play.stacklist = new int [MAX_NOTE_STACK];
	play.stacklist[0] = CLEAR_STACK_NOTE; 
	play.sustain = false;
	play.echo = false;
	
	// The values in stackEdit are constant, and define how note-on events are 
	// interpreted during stack editing, ie, sustained, no stack, polyphonic,
	// and echoing note-on events to the editor
	
	stackEdit.polyphonic = true;	
	stackEdit.stacklist = new int [MAX_NOTE_STACK];
	stackEdit.stacklist[0] = CLEAR_STACK_NOTE; 
	stackEdit.sustain = true;
	stackEdit.echo = true;
	stackEdit.nStacked = 1;
	
	context = &play; 	// changed to &stackEdit for stack editing
	contextPending = 0;	// used to ensure context is only updated synchronously on a new note
	
	machineIndex = 0;
	nMachines = 1;

	/* Get pointers to sound parameters that are accessed during voice allocation */
	stack = paras->getParaByName("stack");
	polyphonic = paras->getParaByName("polyphony");
	portamento = paras->getParaByName("portamnto");
}

void VoiceListItem::moveVoiceStackBelow(VoiceListItem *here, VoiceListItem *lastVoice) {

	/* Extract the voice stack from its old location */

	VoiceListItem *oldHigh = this->getUp();
	VoiceListItem *oldLow = lastVoice->getDown();

	oldHigh->setDown(oldLow);
	oldLow->setUp(oldHigh);

	/* Insert the voice stack at new location */
	
	VoiceListItem *newHigh = here;
	VoiceListItem *newLow = here->getDown();

	newHigh->setDown(this);
	newLow->setUp(lastVoice);

	this->setUp(newHigh);
	lastVoice->setDown(newLow);
}

void VoiceListItem::initialiseAbove(VoiceListItem *here_, signed char voiceIndex_, signed char VDIF_, signed char STDF_) {

	voiceIndex = voiceIndex_; 

	VDIF = VDIF_; 
	STDF = STDF_;

	note = -1; 
       	
	insertAbove(here_);
}

void EntityVoiceAssigner::trace() {

	int a, b;
	
	NoteListItem *n;
	
	/* Count number of entries in notes-on */
	
	a = 0;
	n = topOfNotesOn->getDown();
	while(n != bottomOfNotesOn && a < N_MIDI_NOTES) {
	
		n = n->getDown();
		a++;
	}

	b = 0;
	n = bottomOfNotesOn->getUp();
	while(n != topOfNotesOn && b < N_MIDI_NOTES) {
	
		n = n->getUp();
		b++;
	}

	if(a == b) TRACE("\n%d notes on\t", a);
	
	else { 	TRACE("ERROR: notes-on has bad links\n");
		sound = 0; // disable subsequent note events
		return;
	}
	
	/* Count number of entries in notes-off */
	
	a = 0;
	n = topOfNotesOff->getDown();
	while(n != bottomOfNotesOff && a < N_MIDI_NOTES) {
	
		n = n->getDown();
		a++;
	}

	b = 0;
	n = bottomOfNotesOff->getUp();
	while(n != topOfNotesOff && b < N_MIDI_NOTES) {
	
		n = n->getUp();
		b++;
	}

	if(a == b) TRACE("%d notes off\n", a);
		
	else {	TRACE("\nERROR: notes-off has bad links\n");
		sound = 0; // disable subsequent note events
		return;
	}
	
	VoiceListItem *v;
	
	/* Contents of voices-on */
	
	TRACE("Voices-on:\t");
	
	a = 0;
	v = topOfVoicesOn->getDown();
	while(v != bottomOfVoicesOn && a < MAX_VOICES) {

		TRACE("#%d\t", v->getVoiceIndex());
		v = v->getDown();
		a++; 
	}

	TRACE("\n");

	/* Verify voices-on by reading from bottom up */

	b = 0;
	v = bottomOfVoicesOn->getUp();
	while(v != topOfVoicesOn && b < MAX_VOICES) {

		v = v->getUp();
		b++; 
	}

	if(a != b) {
		
		TRACE("ERROR: voices-on has bad links\n");
		sound = 0; // disable subsequent note events
		return;
	}
	
	/* Contents of voices-off */
	
	TRACE("Voices-off:\t");
	
	a = 0;
	v = topOfVoicesOff->getDown();
	while(v != bottomOfVoicesOff && a < MAX_VOICES) {

		TRACE("#%d\t", v->getVoiceIndex());
		v = v->getDown();
		a++;
	}

	TRACE("\n");

	/* Verify voices-off */

	b = 0;
	v = bottomOfVoicesOff->getUp();
	while(v != topOfVoicesOff && b < MAX_VOICES) {

		v = v->getUp();
		b++; 
	}
	
	if(a != b) {
		
		TRACE("ERROR: voices-off has bad links\n");
		sound = 0; // disable subsequent note events
		return;
	}
	
	TRACE("\n\n");	
}

