;********************************************************************************
;* E N T I T Y 									*
;********************************************************************************
;* File: dsp/lut.asm								*
;*                                                                         	*
;* Copyright (C) A.R.Hardie-Bick 2004-2005                                      *
;*										*
;* This program is free software; you can redistribute it and/or modify		*
;* it under the terms of the GNU General Public License as published by		*
;* the Free Software Foundation; either version 2 of the License, or (at	*
;* your option) any later version.						*
;*										*
;* This program is distributed in the hope that it will be useful, but		*
;* WITHOUT ANY WARRANTY; without even the implied warranty of			*
;* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
;* General Public License for more details.					*
;*										*
;* You should have received a copy of the GNU General Public License		*
;* along with this program; if not, write to the Free Software			*
;* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
;********************************************************************************

	org	Y:$404000

;--------------------------------------------------------------------------
; MISCELLANEOUS 129 STEP BIPOLAR LUTS
;--------------------------------------------------------------------------
;
; EXP1_LUT  is suitable for portamento, attack, decay, release,
;           and velocity decay.
;
; EXP2_LUT  is suitable for LFO delay.
;
; The characteristics of exponential curves are controlled by 'shape'.
;
;                y
;     +1         i .  -  -     y = 1 - { e^(shape*(1-x) }
;              . I
;           .    I             where {..} signifies normalisation
;         .      I             to a range of 0..+1, and shape is
;        .       I             typically between 5 and 8.
;      0 '-------+-------->x
;       -1       0       +1    1-y is used for attack.
;
;--------------------------------------------------------------------------

; EXPONENTIAL LUTS - - - - - - - - - - - - - - - - - - - - - - - - - - - -

            OPT         NOW         ;Disable warning - 1st entry is limited

shape       SET         5.25
min         SET         @XPN(-shape)
max         SET         @XPN(shape)
range       SET         max-min
input       SET         1.0

            DUP         64
             DC         1.0-(@XPN(input*shape)-min)/range
input        SET        input-1.0/64.0
            ENDM
exp1LUT129
            DUP         65
             DC         1.0-(@XPN(input*shape)-min)/range
input        SET        input-1.0/64.0
            ENDM

shape       SET         7.75
min         SET         @XPN(-shape)
max         SET         @XPN(shape)
range       SET         max-min
input       SET         1.0

            DUP         64
             DC         1.0-(@XPN(input*shape)-min)/range
input        SET        input-1.0/64.0
            ENDM
exp2LUT129
            DUP         65
             DC         1.0-(@XPN(input*shape)-min)/range
input        SET        input-1.0/64.0
            ENDM

; SQUARE ROOT - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

input       SET         -1.0

            DUP         64
             DC         @SQT(input/2.0+0.5)
input        SET        input+1.0/64.0
            ENDM
sqrtLUT129
            DUP         65
             DC         @SQT(input/2.0+0.5)
input        SET        input+1.0/64.0
            ENDM

            OPT         W           ;Re-enable warnings

;--------------------------------------------------------------------------
; PAN POWER CORRECTION CURVE
;--------------------------------------------------------------------------
;
; Pan position should not change the total power from the speakers.
; The following equations are used to define left, right and power
; correction factor (see also PAN.BAS):
;
; To calculate the LUT:             pan = -1..+1
;                                   p   = pan/2+.5: p = 0..+1
;                                   c   = (.5/((1-p)^2+p^2))^.5
;
; To calculate pan in real time:    left  = c(1-p)
;                                   right = cp
;
;--------------------------------------------------------------------------

            OPT         NOW         ;Limiting does occur - do not warn

p           SET         0

            DUP         64
             DC         @SQT(.5/(@POW(1-p,2)+@POW(p,2)))
p            SET        p+1.0/128.0
            ENDM
panLUT129
            DUP         65
             DC         @SQT(.5/(@POW(1-p,2)+@POW(p,2)))
p            SET        p+1.0/128.0
            ENDM

            OPT         W           ;Warnings back on

;--------------------------------------------------------------------------
; OVERFLOW COMPENSATION CURVES
;--------------------------------------------------------------------------
; 
; When combining levels at various gains, it is preferable to avoid 
; saturation, which can easily occur if two or more audio signals
; are combined at uncontrolled gains.
;
; One solution is to derive a correction factor from the sum of
; amplitude factors:
;
;         SUMAMP = MIX1 + MIX2 + MIX3 ... + MIXn
;
; In the simplest solution, a correction factor is given by 1/SUMAMP,
; and the gains can be multiplied by this, i.e.
;
;         MX1 = MIX1.f(SUMAMP)
;         MX2 = MIX2.f(SUMAMP) etc.
;
;         where f(SUMAMP) = 1/SUMAMP or 1 if SUMAMP <= 1
;
;
;                    ^
;                 1  I .....
;                    I      .
;                    I       .
;         f(SUMAMP)  I         .
;                    I            .
;                    I                 .
;                    I                       '
;                 0  +------------------------>   SUMAMP
;                    0     1     2     3     4
;
;
; However, this moves abruptly into the corrective region. Smoothing
; the curve results in the following, hopefully more natural, curve:
;
;
;                    ^
;                 1  I.
;                    I   ' .
;                    I       .
;         f(SUMAMP)  I         .
;                    I            .
;                    I                 .
;                    I                       '
;                 0  +------------------------>   SUMAMP
;                    0     1     2     3     4
;
; This shows a curve for correction of up to four signals that are
; being mixed together. The number should be a power of two. For 
; example, the above curve applies equally to the situation where 
; 2 or 3 signals are to be combined. Furthermore, the number of 
; signals can be extended just by extending the graph out to 1/n,
; where n is the maximum number of signals. The only issue then 
; being the precision of the LUT.
;
; Samples for this shape are generated in accordance with
; the following steps:
;
;  for x = 0 to n
;    y = cos(kx)                  ;where k = pi/5.6
;    if x > 1.55 then y = 1/x
;  next x
;
; This is done in SUMAMP.BAS, which can generate suitable 
; numerical data for inclusion below.

sum4LUT65  	; LUT for combining up to four signals

  DC	$7FFFFF,$7FEBDB,$7FAF76,$7F4AE2,$7EBE40,$7E09BC
  DC	$7D2D8E,$7C29FB,$7AFF56,$79ADFC,$783658,$7698DF
  DC	$74D614,$72EE85,$70E2CB,$6EB38A,$6C6174,$69ED43
  DC	$6757BC,$64A1B0,$61CBF9,$5ED77C,$5BC526,$5895EF
  DC	$554AD8,$51EB84,$4EC4EC,$4BDA12,$492492,$469EE5
  DC	$444444,$421084,$400000,$3E0F84,$3C3C3C,$3A83A8
  DC	$38E38E,$3759F2,$35E50D,$348348,$333333,$31F382
  DC	$30C30C,$2FA0BE,$2E8BA3,$2D82D8,$2C8590,$2B9310
  DC	$2AAAAA,$29CBC1,$28F5C2,$282828,$276276,$26A43A
  DC	$25ED09,$253C82,$249249,$23EE09,$234F72,$22B63C
  DC 	$222222,$2192E2,$210842,$208208,$200000
	
;--------------------------------------------------------------------------
; NOISE PROBABILITY DENSITY FUNCTION
;--------------------------------------------------------------------------
; Gives a slightly spikier version of the normal gaussian distribution.
; This LUT is used to transform uniform PDF noise into shaped PDF noise.
;--------------------------------------------------------------------------

  DC	$7FFFF8,$638518,$5C28F0,$575C24,$5428F1,$5147AA
  DC	$4F0A39,$4D1EB4,$4B332F,$499996,$47FFFC,$46B84E
  DC	$4570A0,$4428F2,$433330,$423D6D,$40F5BF,$3FFFFC
  DC	$3F5C25,$3E6663,$3D70A0,$3CCCC9,$3BD707,$3B3330
  DC	$3A3D6D,$399996,$38F5BF,$3851E8,$37AE11,$370A3A
  DC	$366663,$35C28C,$351EB5,$347ADE,$33D707,$33851C
  DC	$32E145,$323D6E,$319997,$3147AB,$30A3D4,$2FFFFE
  DC	$2FAE12,$2F0A3B,$2EB850,$2E1478,$2DC28D,$2D1EB6
  DC	$2CCCCA,$2C28F4,$2BD708,$2B3331,$2AE146,$2A3D6E
  DC	$29EB83,$2947AC,$28F5C0,$28A3D5,$27FFFE,$27AE12  
  DC	$275C27,$26B850,$266664,$261479

pdfLUT129

  DC	$2570A2,$251EB6,$24CCCB,$2428F4,$23D708,$23851D
  DC	$22E146,$228F5A,$223D6F,$219998,$2147AC,$20F5C1
  DC	$20A3D6,$1FFFFE,$1FAE13,$1F5C27,$1EB850,$1E6665
  DC	$1E1479,$1DC28E,$1D1EB7,$1CCCCB,$1C7AE0,$1BD709
  DC	$1B851D,$1B3332,$1A8F5B,$1A3D6F,$19EB84,$1947AD
  DC	$18F5C1,$18A3D6,$17FFFF,$17AE14,$175C28,$16B851
  DC	$166666,$15C28E,$1570A3,$14CCCC,$147AE0,$13D70A
  DC	$13851E,$12E147,$128F5C,$11EB84,$119999,$10F5C2
  DC	$1051EB,$0FAE14,$0F5C28,$0EB851,$0E147A,$0D70A3
  DC	$0CCCCC,$0C28F5,$0B3333,$0A8F5C,$09999A,$08F5C2
  DC	$080000,$06B852,$0570A4,$03D70A,$022C7A

;--------------------------------------------------------------------------
; OVERDRIVE TO FILTER INPUT LEVEL
;--------------------------------------------------------------------------
; This transfer function is a manually designed curve configured to
; be steep at the start and then increasingly flat. This makes it 
; possible to vary the filter input level down to zero, while having 
; a slightly increasing input level through the rest of the range, so 
; distortion can be changed with only a small corresponding change in 
; signal level:
;
; 	f(x) = 1-(1-(x+1-(1-x)^5))^1.87  
;
; See java/distortion/Distortion.java for graph  
;
;               ^
;            1  I           -   '
;               I     . '            
;               I   .           
;         f(x)  I  
;               I .
;               I       
;               I.                      
;            0  +-----------------> x
;               0                1
;
;--------------------------------------------------------------------------

fltOvLUT129

	OPT	NOW	; Do not warn about any limiting

ov	SET	0

	DUP	129
	DC	1.0-@POW(1-0.5*(ov+1.0-@POW(1.0-ov,5)),1.87)
ov	SET	ov+1.0/128.0
	ENDM

	OPT	W	; Warnings back on

