// File: Oscillator.java

// Program to explore oscillator waveforms

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.*;
import javax.swing.event.*;

public class Oscillator extends JComponent 

	implements ActionListener, ChangeListener {

	int sliderSteps = 100;
	
	int freq;
	int feedback;
	int phase;
	
	boolean saw;
	boolean square;

	JSlider freqSlider;
	JSlider feedbackSlider;
	JSlider phaseSlider;

	JPanel controls;
	JPanel sliderPanel;
	JPanel waveformPanel;
	
	JRadioButton b1, b2;
	
	enum Waveform { saw, square };
	Waveform waveform;
	
	public void paint(Graphics g) {

		int x = 0; int y = 0;
		int oldx = 0; int oldy = 0;
		
		int wx = getSize().width;
		int cy = getSize().height / 2;

		g.translate(0, cy);
		
		// Draw axis

		g.drawLine(0, 0, wx, 0);

		// Draw waveform 

		g.setColor(Color.blue);
		
		boolean start = true;

		double F = (double)freq * 0.4 / (sliderSteps * 2.0 * 3.141592638);
		double FB = (double)feedback * 1.3 / sliderSteps;
		double PHZ = (double)phase * 3.141592638 * 2.0 / sliderSteps;

		double Xa = 0, Ya = 0;
		double Xb = 0, Yb = 0;
	
		for(int i = 0; i < wx; i++) {

			// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
			
			// Synths with rich textures typically have two or three oscillators per 
			// voice plus stereo chorus. The chorus effectively multiplies the number 
			// of pitch components by three, although the texture is not as rich as 
			// using the equivalent number of oscillators. The aim with the oscillator
			// design for the Entity Polysynth is to provide two pitch components per
			// oscillator. This turns out to be reasonably code-efficient, and also the
			// parameters controlling each oscillator (pitch, brightness, volume, pan)
			// are only computed (and interpolated) once per pitch component pair.
			// From the user's point of view, a phase parameter is provided. If this is 
			// modulated by an LFO, the effect is the same as having two identical 
			// oscillators detuned from each other (for sawtooth; the square wave 
			// setting is not as strong). This maximises the texture for a given amount 
			// of processing power, without compromising the richness of the sound. 

			// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
			
			// Each oscillator seen by the user is in fact an oscillator pair:

			// First oscillator in pair (Oxa in dsp code)

			Xa = F * i;		// Calculate base phase
			
			if(square) Xa *= 2;	// Double frequency for square wave modulator
			
			Xa -= PHZ;		// Subtract phase offset	

			Xa += FB * Ya;		// Phase modulation by previous output
			
			Ya = Math.sin(Xa);	// Waveform lookup

			// Second oscillator in pair (Oxb in dsp code)

			Xb = F * i;

			Xb += PHZ;			// Add phase offset	

			if(saw) Xb += FB * Yb;	    	// Phase modulation by previous output
			if(square) Xb += FB * Ya;   	// Or phase modulation by first osc.
			
			Yb = Math.sin(Xb+PHZ);		// Waveform lookup
	
			// Combine outputs 
			
			double Y = 0;

			if(saw) Y = (Ya + Yb) * 0.5;	// Mix both outputs together...
			if(square) Y = Yb;		// ...or just use second output

			// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
			
			x = i; y = (int) (Y * 0.45 * (double)cy + 0.5);
			if(!start) g.drawLine(oldx, -oldy, x, -y);
			oldx = x; oldy = y; start = false;

			// Notes - I did try putting an LPF in the feedback loops (just average of 
			// two samples), to see if that sharpens the waveshape before instability
			// sets in. First observation: You can now double the feedback for the
			// corresponding instability threshold. In terms of waveshape, the 
			// instability has moved from the smoothest part of the waveform to the
			// sharpest cusp. Not known if this sounds different - maybe only the 
			// relative phases of harmonics have changed. The waveform does look a 
		        // bit sharper - so maybe it is an improvement. 
	
		}

		controls.repaint();
	}

	public void stateChanged(ChangeEvent e) {

		if(e.getSource() == freqSlider) { 

			freq = freqSlider.getValue();
			repaint();
		}
		
		if(e.getSource() == feedbackSlider) { 

			feedback = feedbackSlider.getValue();
			repaint();
		}
		
		if(e.getSource() == phaseSlider) { 

			phase = phaseSlider.getValue();
			repaint();
		}
	}
	
	public void actionPerformed(ActionEvent e) {

		if ("Saw".equals(e.getActionCommand())) {
			
			saw = true;
			square = false;
			repaint();
		}

		if ("Square".equals(e.getActionCommand())) {
			
			saw = false;
			square = true;
			repaint();
		}
	}

	public Oscillator() { 
	
		freq = 50;
		feedback = 75;
		phase = 1;

		saw = true;
		square = false;
		
		waveform = Waveform.saw;

		controls = new JPanel(new GridBagLayout());
		GridBagConstraints constraints = new GridBagConstraints();
		
		sliderPanel = new JPanel(new GridLayout(2,3));

		freqSlider = new JSlider(JSlider.VERTICAL, sliderSteps, freq);
		feedbackSlider = new JSlider(JSlider.VERTICAL, sliderSteps, feedback);
		phaseSlider = new JSlider(JSlider.VERTICAL, sliderSteps, phase);

		sliderPanel.add(new JLabel("Frequency"));
		sliderPanel.add(new JLabel("Feedback"));
		sliderPanel.add(new JLabel("Phase"));

		sliderPanel.add(freqSlider);
		sliderPanel.add(feedbackSlider);
		sliderPanel.add(phaseSlider);

		waveformPanel = new JPanel();	
		ButtonGroup waveformGroup = new ButtonGroup();

		waveformPanel.add(b1 = new JRadioButton("Saw", saw));		
		waveformGroup.add(b1);
		b1.setActionCommand("Saw");
		b1.addActionListener(this);
		
		waveformPanel.add(b2 = new JRadioButton("Square", square));		
		waveformGroup.add(b2);		
		b2.setActionCommand("Square");
		b2.addActionListener(this);

		constraints.weightx = 7;
		constraints.anchor = GridBagConstraints.WEST;
		controls.add(sliderPanel);
		constraints.weightx = 2;
		constraints.anchor = GridBagConstraints.EAST;
		controls.add(waveformPanel);
		
		setLayout(new FlowLayout());	
		add(controls);
	
		freqSlider.addChangeListener(this);
		feedbackSlider.addChangeListener(this);
		phaseSlider.addChangeListener(this);

	}
		
	public static void main(String[] args) {

		JFrame frame = new JFrame("Entity Poly Oscillator Waveform Explorer");
		Container c = frame.getContentPane();

		c.add(new Oscillator());

		frame.setSize(1000, 700);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setVisible(true);

	}
}
