// File: Switching.java

// Program to illustrate/test translation of continuous values
// supplied by the modulation matrix into switched values.

public class Switching {

	public static void main(String[] args) {

		// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// Switched user parameters, such as octave or LFO waveform, are supplied to the
		// matrix as if they are continuous values, so that they may then be modulated.
		// The full range of switched values, for example 0 to 7, is mapped onto the 
		// full numerical range of a continuously variable matrix input signal.
		// After matrix computation, a (possibly modulated) value is then translated
		// back into its original discrete switched values (0 to 7), which are then
		// applied appropriately to selecting an LFO waveform, or whatever function
		// is appropriate. In order to achieve this, two translations are performed:
		// translation from the user value (0 to 7) to a "continuous" value used
		// for input to the modulation matrix, followed by a translation from the
		// matrix output back into discrete values. The latter is considered first,
		// because this is critical: translation to discrete values outside the 
		// expected range might even crash the DSP, because range checking cannot
		// be performed in a time-critical section of code.
	
		// Define range of values from matrix
			
		double accLimit = 8388608;		// 56K data max
		double accRange = accLimit * 2;		// (signed data range)

		// Define range of desired output values

		double vMin = -10;
		double vMax = +10;

		int	iu;	// integer input	
		
		double 	u;	// user value
		double	m;	// input to matrix
		double 	x;	// input from matrix
		double	y;	// output
		
		int 	iy;	// integer output
		
		double qp, k;
		
		// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// The idea is that the next values outside the range should be just beyond what's 
		// possible for the accLimit value. In other words, for a matrix output of accLimit,
		// say 8388608 (assume for the sake of argument that it's not 8388607), you want
		// to know for *absolute* certain that this will not give an output greater than 
		// vMax.
		//
		// In the simplest instance, we know that:
		//
		// 	vMax + 1 > accLimit * q
		//
		// So we want to know for what values of q this is true:
		//
		//	       vMax + 1
		// 	q  <  ----------
		// 	       accLimit
		//
		// For example, if vMax = 10:
		//
		// 	q  <  1.311 E-6
		//
		// Multiply this by 65536 to get the factor used in the code 
		// (where it is then right shifted 16 places):
		//
		// 	q' < 0.0859375
		// 	q' = 0.0859374
		//
		// For an arbitrary range of output values, we get:
		//
		// 	      (vMax - vMin) + 1
		// 	q  <  -----------------
		//	          accRange
		// 
		// or
		// 	      (vMax - vMin) + 1
		// 	q' =  ----------------- . 65536 - 0.0001
		//	          accRange
		//	          
		// and an offset needs to be added to the final result:
		//
		// 	k = (vMax + vMin) / 2
		//
		// so the translation from matrix output, x, to discrete 
		// values is:
		//
		// 	y = (xq' >> 16) + k
		//
		// (and add 0.5 to the result before truncating to integer)
		// - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
		// EXAMPLE
		
		vMin = -10; vMax = +10;

		System.out.printf("\nvMin = %d\tvMax = %d\n", (int)vMin, (int)vMax);
		
		qp = (((vMax - vMin) + 1) / accRange ) * 65536.0 - 0.0001;
		k  = (vMax + vMin) / 2.0;

		x = -8388608; 
		y = (x * qp / 65536) + k + 0.5; 
		
		iy = tcInt(y);
		System.out.printf("%f\t->\t%f\t%d\n", x, y, iy);

		x = 8388607; 
		y = (x * qp / 65536) + k + 0.5;
		
		iy = tcInt(y);
		System.out.printf("%f\t->\t%f\t%d\n", x, y, iy);

		// EXAMPLE 
		
		vMin = 0; vMax = +5;

		System.out.printf("\nvMin = %d\tvMax = %d\n", (int)vMin, (int)vMax);
		
		qp = (((vMax - vMin) + 1) / accRange ) * 65536.0 - 0.0001;
		k  = (vMax + vMin) / 2.0;

		x = -8388608; 
		y = (x * qp / 65536) + k + 0.5; 
		
		iy = tcInt(y);
		System.out.printf("%f\t->\t%f\t%d\n", x, y, iy);

		x = 8388607; 
		y = (x * qp / 65536) + k + 0.5; 
		
		iy = tcInt(y);
		System.out.printf("%f\t->\t%f\t%d\n", x, y, iy);

		// EXAMPLE 
		
		vMin = 0; vMax = +35;

		System.out.printf("\nvMin = %d\tvMax = %d\n", (int)vMin, (int)vMax);
		
		qp = (((vMax - vMin) + 1) / accRange ) * 65536.0 - 0.0001;
		k  = (vMax + vMin) / 2.0;

		x = -8388608; 
		y = (x * qp / 65536) + k + 0.5; 
		
		iy = tcInt(y);
		System.out.printf("%f\t->\t%f\t%d\n", x, y, iy);

		x = 8388607; 
		y = (x * qp / 65536) + k + 0.5; 
		
		iy = tcInt(y);
		System.out.printf("%f\t->\t%f\t%d\n", x, y, iy);

		System.out.printf("\n");
		
		// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// The user value (when unmodulated) preferably sits halfway between two switch
		// thresholds. We need to translate the user values, such as -10 to +10, into
		// larger values for the matrix input.
		//
		// The matrix input value, m, is calculated from the user value u, as:
		//
		// 	        u - k 	        
		// 	m  =  ----------   (by rearranging the equation from above)	
		// 	       q' >> 16    
		//
		// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

		// EXAMPLE

		vMin = -10; vMax = 10;

		qp = (((vMax - vMin) + 1) / accRange ) * 65536.0 - 0.0001;
		k  = (vMax + vMin) / 2.0;

		for(iu = -10; iu <= 10; iu++) {
		
			u = (double)iu;			// user value
			m = (u-k) / (qp / 65536);	// matrix input
			x = m;				// matrix processing
			y = (x * qp / 65536) + k + 0.5;	// retranslate

			iy = tcInt(y);					

			System.out.printf("%d\t%f\t->\t%f\t%d\n", iu, x, y, iy);
		}

		System.out.printf("\n");
		
		// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// SUMMARY OF EQUATIONS
		// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
		// Translating a user integer value, u, into a matrix input value, m:
		// (parameter translation, as performed in EntitySound.cpp)
		//
		// 	        u - k 	        
		// 	m  =  ---------- 		(Equation #1)
		// 	       q' >> 16    
		//
		// Translating from matrix output, x, back to the same integer range:
		// (matrix output translation, as performed in dsp/parameterGenerator.asm)
		//
		// 	y = (xq' >> 16) + k + 0.5  	(Equation #2)
		//
		// (then discard the fractional part of the 2's complement result)
		// 	
		// Deriving the constants q' and k used in the two equations above:
		// 
		//					(Equation #3)
		// 	      (vMax - vMin) + 1
		// 	q' =  ----------------- . 65536 - 0.0001
		//	          accRange
		//	          
		// 	k = (vMax + vMin) / 2		(Equation #4)
		//
		// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		// - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
		// EXAMPLES
		// 
		// Sound octave -5..+5, k = 0,  qp = 0.04286875  
		// Osc octave -10..+10, k = 0,  qp = 0.08193125
		// Mod source 0..8,     k = 4,  qp = 0.03115
	        // LFO waveform 0..12,	k = 6,  qp = 0.05068125
		// LFO sync 0..26,	k = 13, qp = 0.10556875
		// - - - - - - - - - - - - - - - - - - - - - - - -
	}
	
	static int tcInt(double f) {

		// Truncate a signed fraction to integer
		// as occurs in two's complement binary
		
		if(f < 0.0) return (int)(f - 1.0);
		return (int)f;
	}	
}	

