/********************************************************************************
 * U T I L S									*
 ********************************************************************************
 * File: utils.cpp								*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/

#include "utils.h"

//*******************************************************************************
//
//		rshift - a simple random number generator
//
//*******************************************************************************
// Random number generator - very basic, generates random words suitable for
// seeding an r250. It has as its sole advantage that it has only one word 
// of initial state.

int rshift::rn() {	// Private method to generate a bad word
			// (low order bits are very predictable)
	
	return (state = state * 1664525 + 1013904223);
}

int rshift::rng() {	// This is the public method that generates a better
			// quality word by combining the high order
			// bits from two iterations of rn()

	return ((rn() & 0xFFFF0000) | ((rn() & 0xFFFF0000) >> 16));	
}

rshift::rshift(unsigned int seed = MAGIC1) {
	state = seed;
}

//*******************************************************************************
//
//		r250 - a fast good quality random number generator
//
//*******************************************************************************

int r250::getpos() { return si; }

void r250::setpos(int n) {

	si = (unsigned char) n;
	n += 103;
	if(n > 249) n -= 250;
	sj = (unsigned char) n;
}		

// * Manually seed r250 from a single value *

void r250::seed(int seed = MAGIC1) {

	rshift rx(seed); for(int x = 0; x < 250; x++) state[x] = rx.rng(); 
	
	validate(); 
	
	si = 0; sj = 103;

}

// * Manually seed r250 from an array of values *

void r250::seed(int* seed) {

	for(int x = 0; x < 250; x++) state[x] = seed[x];
	
	validate(); 
	
	si = 0; sj = 103;
}

// * Iterate n times *

void r250::spin(int n = 1) {

	if(n > 0) for(int x = 1; x <= n; x++) rng();
	if(n < 0) for(int x = -1; x >= n; x--) gnr();	
}
	
void r250::validate() {

// Ensure columns are non-zero and different * * * * * * * * * * * * * * * * * * * * * * * * * * * 

// Note - it is important that any modifications here are repeatable, as the r250 may be used
// for a repeatable sequence based on known starting conditions. The code here achieves this.
// Validation is a non-ambiguous function of the r250 state, enforcing certain conditions.

rshift rx(MAGIC2);		// A source for pseudo-randomly modifying slices

int slice[8][32];		// Eight words per slice, thirty-two slices

int i, j, k, x, z;

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// First copy r250 state into slices of 250 bits, arranged as 8 x 32 bit words

	for(i = 0; i < 32; i++) {				// Bit slice counter
		for(x = 0; x < 8; x++) slice[x][i] = 0;		// Initialise slice to all zeros
		
		for(j = 0; j < 250; j++)			// Bit counter in destination slice
			if(state[j] & mask[i])			// If state bit is set
				slice[j>>5][i] |= mask[j&31];	// set corresponding bit in slice
	}
		
// Check for all zeros in any of the columns (incredibly unlikely)

	for(i = 0; i < 32; i++) {
		z = 1;
		for(j = 0; j < 8; j++)	if(slice[j][i]) z = 0;	// If a word is non zero, slice is okay
	 	if(z) fix_slice(slice, i, rx);			// If all are zero, flip some bits
	}
	
// Check for matching columns (again, incredibly unlikely)

x = 1; while(x) {
	x = 0; for(i = 0; i < 32; i++)					// Select slice i
		for(j = 0; j < 32; j++)					// Select slice j
			if(i != j) {					// Compare only different slices
				z = 1;					// Assume dodgy
				for(k = 0; k < 8; k++)			// Select each word in slice
	
					if(slice[k][i] != slice[k][j]) z = 0;	// If different, slice okay
				if(z) { 
					fix_slice(slice, j, rx); x = 1;		// If slices match, change one
				}						// and ensure whole sequence 
			}							// repeats again
		}

}

// Flip some bits in a slice, read back slices into r250 state

void r250::fix_slice(int slice[8][32], int i, rshift &rx) {

int j, x, z;

	for(x = 0; x < 100; x++) {				// Select 100 bits at random and flip them
		while((z = rx.rng() & 0xff) > 249);		// Generate bit address < 250
		slice[z>>5][i] ^= mask[z&31];			// Flip bit at that address
	}
		
// Read back slices into state

	for(x = 0; x < 250; x++) state[x] = 0;			// Initialise state to all zeros

	for(i = 0; i < 32; i++) {				// Slice counter
		for(j = 0; j < 250; j++)			// Bit counter in source slice
			if((slice[j>>5][i] & mask[j&31]))	// If slice bit is set
				state[j] |= mask[i];		// set corresponding bit in state
	}
	
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// * Constructor defines r250 default state and pointer offsets *

r250::r250(int seed = MAGIC1) {

// Initialise incr and decr luts for r250

	for(int x = 0; x < 249; x++) incr[x] = x+1; 
	incr[249] = 0;
	
	for(int x = 249; x > 0; x--) decr[x] = x-1;
	decr[0] = 249;

// Initialise bit mask luts - also used by constructor in r250::validate()

	mask[0] = 1;
	for(int x = 1; x < 32; x++) mask[x] = 1 << x;

		// mask[0] = ...0001
		// mask[1] = ...0010
		// mask[2] = ...0100
		// mask[3] = ...1000
		// ...

// State initialisation and validation - by default fill the state using a simple rng 
	
	si = 0; sj = 103;	// Counter offsets into state rows

	rshift rx(seed); for(int x = 0; x < 250; x++) state[x] = rx.rng();

	validate();		// Check columns of default state
}

//*******************************************************************************
//
//		BTree - A simple binary tree for text
//
//*******************************************************************************

void *BTree::lookup(char *text) {

	switch(compare(text)) {

		case 0  : 	return data;
				
		case +1 : 	if(!high) return 0; 
				return high->lookup(text);
				
		case -1 : 	if(!low) return 0; 
				return low->lookup(text);
	}

	return 0;
}

bool BTree::add(char *newKey, void *newData) {

	switch(compare(newKey)) {

		case 0  : 	return false;
		
		case +1 : 	if(!high) {
				  
					high = new BTree(newKey, newData);
				  	return true;
			  	}

			  	else return high->add(newKey, newData);
			  
		case -1 : 	if(!low) {
				  
					low = new BTree(newKey, newData);
				  	return true;
			  	}

				else return low->add(newKey, newData);
	}
	
	return false;
}

int BTree::size() {

	int n = 1;

	if(high) n += high->size();
	if(low) n += low->size();
	
	return n;
}

int BTree::compare(char *t) {

	char *k = key;
	
	while(*k == *t) {

		if(!*k) return 0; 
		k++; t++;
	}

	return (*k - *t) > 0 ? 1 : -1;
}

BTree::BTree(char *newKey, void *newData) { 		

	int x = 0; char *k = newKey;
	while(*k++) x++;
	
	key = new char [x];

	char *a = newKey;
	char *b = key;

	while((*b++ = *a++));
	
	data = newData;
	
	low = 0;
	high = 0;			
}

BTree::~BTree() {

	if(high) delete high;
	if(low) delete low;
	
	delete [] key;
}

//*******************************************************************************
//
//		Hash - A simple hash table for text
//
//*******************************************************************************

void *Hash::lookup(char *text) {

	int hashKey = hashFunction(text);
	BTree *node = table[hashKey];
	
	if(!node) return 0;

	else return node->lookup(text);
}

bool Hash::add(char *newKey, void *newData) {

	int hashKey = hashFunction(newKey);
	BTree *node = table[hashKey];

	if(!node) {
		
		table[hashKey] = new BTree(newKey, newData);
		return true;
	}

	return node->add(newKey, newData);
}

int Hash::hashFunction(char *key) {

	int index, hash = 0;
	
	while(*key) {
		
		index = hash + (unsigned char) *key++;
		hash = rTable[index & 255] ^ (hash << 1);
	}

	return hash & (tableSize - 1);
}

void Hash::getStats(int &direct, int &total) {
	
	direct = 0; 
	total = 0;
	
	for(int x = 0; x < tableSize; x++) {
		
		if(table[x]) {
			
			direct++;

			total += table[x]->size();
		}
	}
}

Hash::Hash(int hTableSize) {

	tableSize = 1; 

	while(tableSize < hTableSize) tableSize <<= 1;

	r250 r;	
	
	for(int x = 0; x < 256; x++) {
		
		rTable[x] = r.rng();
	}

	table = new BTree* [tableSize];

	for(int x = 0; x < tableSize; x++) {
		
		table[x] = 0;
	}
}

Hash::~Hash() {

	for(int x = 0; x < tableSize; x++) {

		if(table[x]) delete table[x];
	}

	delete [] table;
}

//*******************************************************************************
//
// CCITT32 CRC checksum algorithm (from page 447 of The Data Compression Book)
//
//*******************************************************************************

CRC32::CRC32() {

	for(int i = 0; i < 256; i++) {

		unsigned int j = i;

		for(int k = 0; k < 8; k++) {

			if(j & 1) j = (j >> 1 ) ^ 0xEDB88320L;	// CCITT32 CRC Polynomial
			
			else	  j >>= 1;
		}
		
		crcLUT[i] = j;
	}
}

unsigned int CRC32::checksum(void *buffer, unsigned int nBytes) {
			
	unsigned char *p = (unsigned char*)buffer;
			
	unsigned int crc = 0xffffffff;

	while(nBytes--)	{
		
		crc = (crc >> 8) ^ crcLUT[(crc ^ *p++) & 0xff];
	}

	return crc;
}

//*******************************************************************************
//
// Bitstream class for reading and writing bits in a buffer
//
//*******************************************************************************

void BitStream::init(void *buffer) {

	ptr = (unsigned char*) buffer;
	mask = 128;
	byte = 0;
}

int BitStream::getnBytes() {

	return mask == 128 ? byte : byte + 1;
}

int BitStream::get(int nBits) {

	int value = 0;
	
	while(nBits--) {

		value |= mask & *ptr ? 1 << nBits : 0;	/* MS bit is first */					
		if(!(mask >>= 1)) { mask = 128; ptr++; byte++; }
	}

	return value;
}

void BitStream::put(int value, int nBits) {

	if(mask == 128) *ptr = 0;

	while(nBits--) {

		*ptr |= (value >> nBits) & 1 ? mask : 0;
		if(!(mask >>= 1)) { mask = 128; *++ptr = 0; byte++; }
	}
}

//*******************************************************************************
//
// Fast integer to string conversion
//
//*******************************************************************************

char *i2s(int i) { // Signed integer to four digit string conversion
	
	static 	char number[32];
	
	int 	s, x, y;
	char 	*n;
	bool 	d;
	
	d = true;	// Still on first digit?
	n = number;
	
	if(!i) {

		for(x = 0; x < 4; x++) *n++ = 32;
			
		*n++ = '0';
		*n++ = 0;

		return number;
	}		

	if(i < 0) {

		s = '-';
		i = ~i + 1;
	} 
	
	else s = ' ';

	for(x = 1000; (x); x /= 10) {
	
		y = i / x;

		if(y) {
	
			i = i % x;
	
			if(d) {
				*n++ = s;
				d = false;
			} 	
			
			*n++ = y + '0';	
			
		} else 	d ? *n++ = ' ' : *n++ = '0';
	}

	if(d) {
		*n++ = ' ';
		*n++ = '0';
	}

	*n++ = 0;
	
	return number;
}
	












